import csv
import datetime as dt

# Set up logging with a level of WARNING
import os
from collections import defaultdict
from dataclasses import fields

import numpy as np

from pythermalcomfort.classes_return import (
    JOS3BodyParts,
    JOS3Output,
)
from pythermalcomfort.jos3_functions import construction as cons
from pythermalcomfort.jos3_functions import matrix
from pythermalcomfort.jos3_functions import thermoregulation as threg
from pythermalcomfort.jos3_functions.construction import (
    pass_values_to_jos3_body_parts,
    to_array_body_parts,
    validate_body_parameters,
)
from pythermalcomfort.jos3_functions.matrix import (
    INDEX,
    NUM_NODES,
    VINDEX,
    remove_body_name,
)
from pythermalcomfort.jos3_functions.parameters import ALL_OUT_PARAMS, Default
from pythermalcomfort.models.pmv_ppd_iso import pmv_ppd_iso
from pythermalcomfort.utilities import Models, Postures, antoine, met_to_w_m2


class JOS3:
    """JOS-3 model simulates human thermal physiology including skin temperature, core
    temperature, sweating rate, etc. for the whole body and 17 local body parts.

    This model was developed at Shin-ichi Tanabe Laboratory, Waseda University
    and was derived from 65 Multi-Node model (https://doi.org/10.1016/S0378-7788(02)00014-2)
    and JOS-2 model (https://doi.org/10.1016/j.buildenv.2013.04.013).

    To use this model, create an instance of the JOS3 class with optional body parameters
    such as body height, weight, age, sex, etc.

    Environmental conditions such as air temperature, mean radiant temperature, air velocity, etc.
    can be set using the setter methods. (ex. X.tdb, X.tr, X.v)
    If you want to set the different conditions in each body part, set them
    as a 17 lengths of list, dictionary, or numpy array format.

    List or numpy array format input must be 17 lengths and means the order of "head", "neck", "chest",
    "back", "pelvis", "left_shoulder", "left_arm", "left_hand", "right_shoulder", "right_arm",
    "right_hand", "left_thigh", "left_leg", "left_foot", "right_thigh", "right_leg" and "right_foot".

    The model output includes local and mean skin temperature, local core temperature,
    local and mean skin wettedness, and heat loss from the skin etc.
    The model output can be accessed using the `dict_results()` method and be converted to a csv file
    using the `to_csv` method.
    Each output parameter also can be accessed using getter methods.
    (ex. X.t_skin, X.t_skin_mean, X.t_core)

    If you use this package, please cite us as follows and mention the version of pythermalcomfort used:
    Y. Takahashi, A. Nomoto, S. Yoda, R. Hisayama, M. Ogata, Y. Ozeki, S. Tanabe,
    Thermoregulation Model JOS-3 with New Open Source Code, Energy & Buildings (2020),
    doi: https://doi.org/10.1016/j.enbuild.2020.110575

    Note: To maintain consistency in variable names for pythermalcomfort,
    some variable names differ from those used in the original paper.

    Attributes
    ----------
    tdb : float or list of floats
        Dry bulb air temperature.
    tr : float or list of floats
        Mean radiant temperature.
    to : float or list of floats
        Operative temperature.
    rh : float or list of floats
        Relative humidity.
    v : float or list of floats
        Air velocity.
    clo : float or list of floats
        Clothing insulation.
    posture : str
        Posture of the subject.
    par : float
        Physical activity ratio.
    t_body : numpy.ndarray
        Body temperature.
    bsa : numpy.ndarray
        Body surface area.
    r_t : numpy.ndarray
        Radiative heat transfer coefficient.
    r_et : numpy.ndarray
        Evaporative heat transfer coefficient.
    w : numpy.ndarray
        Skin wettedness.
    w_mean : float
        Mean skin wettedness.
    t_skin_mean : float
        Mean skin temperature.
    t_skin : numpy.ndarray
        Skin temperature.
    t_core : numpy.ndarray
        Core temperature.
    t_cb : numpy.ndarray
        Central blood temperature.
    t_artery : numpy.ndarray
        Arterial blood temperature.
    t_vein : numpy.ndarray
        Venous blood temperature.
    t_superficial_vein : numpy.ndarray
        Superficial venous blood temperature.
    t_muscle : numpy.ndarray
        Muscle temperature.
    t_fat : numpy.ndarray
        Fat temperature.
    body_names : list of str
        Names of the body parts.
    results : dict
        Dictionary of the simulation results.
    bmr : float
        Basal metabolic rate.
    version : str
        Version of the JOS3 model.

    Returns
    -------
    cardiac_output :
        cardiac output (the sum of the whole blood flow) [L/h]
    dt :
        time step [sec]
    q_res :
        heat loss by respiration [W]
    q_skin2env :
        total heat loss from the skin (each body part) [W]
    q_thermogenesis_total :
        total thermogenesis of the whole body [W]
    simulation_time :
        simulation times [sec]
    t_core :
        core temperature (each body part) [°C]
    t_skin :
        skin temperature (each body part) [°C]
    t_skin_mean :
        mean skin temperature [°C]
    w :
        skin wettedness (each body part) [-]
    w_mean :
        mean skin wettedness [-]
    weight_loss_by_evap_and_res :
        weight loss by the evaporation and respiration of the whole body [g/sec]
    age :
        age [years]
    bf_ava_foot :
        AVA blood flow rate of one foot [L/h]
    bf_ava_hand :
        AVA blood flow rate of one hand [L/h]
    bf_core :
        core blood flow rate (each body part) [L/h]
    bf_fat :
        fat blood flow rate (each body part) [L/h]
    bf_muscle :
        muscle blood flow rate (each body part) [L/h]
    bf_skin :
        skin blood flow rate (each body part) [L/h]
    bsa :
        body surface area (each body part) [m2]
    clo :
        clothing insulation (each body part) [clo]
    e_max :
        maximum evaporative heat loss from the skin (each body part) [W]
    e_skin :
        evaporative heat loss from the skin (each body part) [W]
    e_sweat :
        evaporative heat loss from the skin by only sweating (each body part) [W]
    fat :
        body fat rate [%]
    height :
        body height [m]
    par :
        physical activity ratio [-]
    q_bmr_core :
        core thermogenesis by basal metabolism (each body part) [W]
    q_bmr_fat :
        fat thermogenesis by basal metabolism (each body part) [W]
    q_bmr_muscle :
        muscle thermogenesis by basal metabolism (each body part) [W]
    q_bmr_skin :
        skin thermogenesis by basal metabolism (each body part) [W]
    q_nst :
        core thermogenesis by non-shivering (each body part) [W]
    q_res_latent :
        latent heat loss by respiration (each body part) [W]
    q_res_sensible :
        sensible heat loss by respiration (each body part) [W]
    q_shiv :
        core or muscle thermogenesis by shivering (each body part) [W]
    q_skin2env_latent :
        latent heat loss from the skin (each body part) [W]
    q_skin2env_sensible :
        sensible heat loss from the skin (each body part) [W]
    q_thermogenesis_core :
        core total thermogenesis (each body part) [W]
    q_thermogenesis_fat :
        fat total thermogenesis (each body part) [W]
    q_thermogenesis_muscle :
        muscle total thermogenesis (each body part) [W]
    q_thermogenesis_skin :
        skin total thermogenesis (each body part) [W]
    q_work :
        core or muscle thermogenesis by work (each body part) [W]
    r_et :
        total clothing evaporative heat resistance (each body part) [(m2*kPa)/W]
    r_t :
        total clothing heat resistance (each body part) [(m2*K)/W]
    rh :
        relative humidity (each body part) [%]
    sex : str
        sex
    t_artery :
        arterial temperature (each body part) [°C]
    t_cb :
        central blood temperature [°C]
    t_core_set :
        core set point temperature (each body part) [°C]
    t_fat :
        fat temperature (each body part) [°C]
    t_muscle :
        muscle temperature (each body part) [°C]
    t_skin_set :
        skin set point temperature (each body part) [°C]
    t_superficial_vein :
        superficial vein temperature (each body part) [°C]
    t_vein :
        vein temperature (each body part) [°C]
    tdb :
        dry bulb air temperature (each body part) [°C]
    to :
        operative temperature (each body part) [°C]
    tr :
        mean radiant temperature (each body part) [°C]
    v :
        air velocity (each body part) [m/s]
    weight :
        body weight [kg]

    Examples
    --------
    Build a model and set a body built
    Create an instance of the JOS3 class with optional body parameters such as body height, weight, age, sex, etc.

    .. code-block:: python

        from pythermalcomfort.classes_return import get_attribute_values
        import numpy as np
        import pandas as pd
        import matplotlib.pyplot as plt
        import os
        from pythermalcomfort.models import JOS3
        from pythermalcomfort.jos3_functions.parameters import (
            local_clo_typical_ensembles,
        )

        model = JOS3(
            height=1.7,
            weight=60,
            fat=20,
            age=30,
            sex="male",
            bmr_equation="japanese",
            bsa_equation="fujimoto",
        )
        # Set environmental conditions such as air temperature, mean radiant temperature using the setter methods.
        # Set the first condition
        # Environmental parameters can be input as int, float, list, dict, numpy array format.
        model.tdb = 28  # Air temperature [°C]
        model.tr = 30  # Mean radiant temperature [°C]
        model.rh = 40  # Relative humidity [%]
        model.v = np.array(  # Air velocity [m/s]
            [
                0.2,  # head
                0.4,  # neck
                0.4,  # chest
                0.1,  # back
                0.1,  # pelvis
                0.4,  # left shoulder
                0.4,  # left arm
                0.4,  # left hand
                0.4,  # right shoulder
                0.4,  # right arm
                0.4,  # right hand
                0.1,  # left thigh
                0.1,  # left leg
                0.1,  # left foot
                0.1,  # right thigh
                0.1,  # right leg
                0.1,  # right foot
            ]
        )
        model.clo = get_attribute_values(
            local_clo_typical_ensembles[
                "briefs, socks, undershirt, work jacket, work pants, safety shoes"
            ]["local_body_part"]
        )
        # par should be input as int, float.
        model.par = 1.2  # Physical activity ratio [-], assuming a sitting position
        # posture should be input as int (0, 1, or 2) or str ("standing", "sitting" or "lying").
        # (0="standing", 1="sitting" or 2="lying")
        model.posture = "sitting"  # Posture [-], assuming a sitting position

        # Run JOS-3 model
        model.simulate(
            times=30,  # Number of loops of a simulation
            dtime=60,  # Time delta [sec]. The default is 60.
        )  # Exposure time = 30 [loops] * 60 [sec] = 30 [min]
        # Set the next condition (You only need to change the parameters that you want to change)
        model.to = 20  # Change operative temperature
        model.v = {  # Air velocity [m/s], assuming to use a desk fan
            "head": 0.2,
            "neck": 0.4,
            "chest": 0.4,
            "back": 0.1,
            "pelvis": 0.1,
            "left_shoulder": 0.4,
            "left_arm": 0.4,
            "left_hand": 0.4,
            "right_shoulder": 0.4,
            "right_arm": 0.4,
            "right_hand": 0.4,
            "left_thigh": 0.1,
            "left_leg": 0.1,
            "left_foot": 0.1,
            "right_thigh": 0.1,
            "right_leg": 0.1,
            "right_foot": 0.1,
        }
        # Run JOS-3 model
        model.simulate(
            times=60,  # Number of loops of a simulation
            dtime=60,  # Time delta [sec]. The default is 60.
        )  # Additional exposure time = 60 [loops] * 60 [sec] = 60 [min]
        # Set the next condition (You only need to change the parameters that you want to change)
        model.tdb = 30  # Change air temperature [°C]
        model.tr = 35  # Change mean radiant temperature [°C]
        # Run JOS-3 model
        model.simulate(
            times=30,  # Number of loops of a simulation
            dtime=60,  # Time delta [sec]. The default is 60.
        )  # Additional exposure time = 30 [loops] * 60 [sec] = 30 [min]

        # The easiest way to access the results is to use the `results` method
        results = model.results()
        # you can then use dot notation to access the results
        print(
            results.t_skin_mean
        )  # Print the mean skin temperature or you can use print(results['t_skin_mean'])
        # some attributes have results for each body part, you can access them by using the body part name
        print(results.t_skin.head)  # Print the skin temperature of the head

        # You can also save the results as a pandas dataframe and plot them
        df = pd.DataFrame(
            [results.t_skin.head, results.t_skin.pelvis]
        ).transpose()  # Make pandas.DataFrame
        df.plot()  # Plot time series of local skin temperature.
        plt.legend(["Head", "Pelvis"])  # Reset the legends
        plt.ylabel("Skin temperature [°C]")  # Set y-label as 'Skin temperature [°C]'
        plt.xlabel("Time [min]")  # Set x-label as 'Time [min]'
        plt.show()  # Show the plot

    """

    def __init__(
        self,
        height: float = Default.height,
        weight: float = Default.weight,
        fat: float = Default.body_fat,
        age: int = Default.age,
        sex: str = Default.sex,
        ci: float = Default.cardiac_index,
        bmr_equation: str = Default.bmr_equation,
        bsa_equation: str = Default.bsa_equation,
    ):
        """Initialize a new instance of the JOS3 class, which models and simulates various
        physiological parameters related to human thermoregulation.

        This class uses mathematical models to calculate and predict body temperature,
        basal metabolic rate, body surface area, and other related parameters.

        Parameters
        ----------
        height : float, optional
            Body height in meters.
        weight : float, optional
            Body weight in kilograms.
        fat : float, optional
            Fat percentage.
        age : int, optional
            Age in years.
        sex : str, optional
            Sex ("male" or "female").
        ci : float, optional
            Cardiac index in liters per minute per square meter.
        bmr_equation : str, optional
            The equation used to calculate basal metabolic rate (BMR). Options are "harris-benedict"
            for Caucasian data (DOI: doi.org/10.1073/pnas.4.12.370) or "japanese" for Ganpule's equation
            (DOI: doi.org/10.1038/sj.ejcn.1602645).
        bsa_equation : str, optional
            The equation used to calculate body surface area (BSA). Choose one from pythermalcomfort.utilities.BodySurfaceAreaEquations.

        Returns
        -------
        None.

        Examples
        --------
        Create an instance of the JOS3 class with optional body parameters:

        .. code-block:: python

            jos3_model = JOS3(height=1.75, weight=70, age=25, sex="female")

        """
        # validate body parameters
        validate_body_parameters(height=height, weight=weight, age=age, body_fat=fat)

        # Initialize basic attributes
        self._height = height
        self._weight = weight
        self._fat = fat
        self._sex = sex
        self._age = age
        self._ci = ci
        self._bmr_equation = bmr_equation
        self._bsa_equation = bsa_equation

        # Calculate body surface area (bsa) rate
        self._bsa_rate = cons.bsa_rate(
            height=height,
            weight=weight,
            bsa_equation=bsa_equation,
        )

        # Calculate local body surface area
        self._bsa = cons.local_bsa(
            height=height,
            weight=weight,
            bsa_equation=bsa_equation,
        )

        # Calculate basal blood flow (BFB) rate [-]
        self._bfb_rate = cons.bfb_rate(
            height=height,
            weight=weight,
            bsa_equation=bsa_equation,
            age=age,
            ci=ci,
        )

        # Calculate thermal conductance (CDT) [W/K]
        self._cdt = cons.conductance(
            height=height,
            weight=weight,
            bsa_equation=bsa_equation,
            fat=fat,
        )

        # Calculate thermal capacity [J/K]
        self._cap = cons.capacity(
            height=height,
            weight=weight,
            bsa_equation=bsa_equation,
            age=age,
            ci=ci,
        )

        # Set initial core and skin temperature set points [°C]
        self.cr_set_point = np.ones(Default.num_body_parts) * Default.core_temperature
        self.sk_set_point = np.ones(Default.num_body_parts) * Default.skin_temperature

        # Initialize body temperature [°C]
        self._t_body = np.ones(NUM_NODES) * Default.other_body_temperature

        # Initialize environmental conditions and other factors
        # (Default values of input conditions)
        self._tdb: np.ndarray = (
            np.ones(Default.num_body_parts) * Default.dry_bulb_air_temperature
        )
        self._tr = np.ones(Default.num_body_parts) * Default.mean_radiant_temperature
        self._rh = np.ones(Default.num_body_parts) * Default.relative_humidity
        self._v = np.ones(Default.num_body_parts) * Default.air_speed
        self._clo = np.ones(Default.num_body_parts) * Default.clothing_insulation
        self._iclo = (
            np.ones(Default.num_body_parts)
            * Default.clothing_vapor_permeation_efficiency
        )
        self._par = Default.physical_activity_ratio
        self._posture = Default.posture
        self._hc = None  # Convective heat transfer coefficient
        self._hr = None  # Radiative heat transfer coefficient
        self.ex_q = np.zeros(NUM_NODES)  # External heat gain
        self._time = dt.timedelta(0)  # Elapsed time
        self.model_name = "JOS3"  # Model name
        # TODO expose them as single properties and not as a dict and document them
        self.options = {
            "nonshivering_thermogenesis": True,
            "cold_acclimated": False,
            "shivering_threshold": False,
            "limit_dshiv/dt": False,
            "bat_positive": False,
            "ava_zero": False,
            # TODO shivering is not used in the model
            "shivering": False,
        }

        # Set shivering threshold = 0
        threg.PRE_SHIV = 0

        # Initialize history to store model parameters
        self._history: list[JOS3Output] = []

        # Set elapsed time to 0
        self._time = dt.timedelta(0)  # Elapsed time

        # Reset set-point temperature and save the last model parameters
        dict_results = self._reset_setpt()
        # TODO why the first element of the simulation is for a naked person?
        self._history.append(dict_results)

    def _calculate_operative_temp_when_pmv_is_zero(
        self,
        v: float,
        rh: float,
        met: float,
        clo: float,
    ) -> float:
        """Calculate operative temperature [°C] when PMV=0 with NaN handling and retry
        logic.

        Parameters
        ----------
        v : float, optional
            Air velocity [m/s]. The default is 0.1.
        rh : float, optional
            Relative humidity [%]. The default is 50.
        met : float, optional
            Metabolic rate [met]. The default is 1.
        clo : float, optional
            Clothing insulation [clo]. The default is 0.

        Returns
        -------
        to : float
            Operative temperature [°C].

        """
        # Default parameters
        initial_to = 28
        tolerance = 0.001
        max_iterations = 100
        adjustment_factor = 3
        retry_adjustment_factor = adjustment_factor * 200
        retry_attempts = 100

        to = initial_to

        # Main loop for finding PMV=0
        for _i in range(max_iterations):
            pmv_value = pmv_ppd_iso(
                to,
                to,
                v,
                rh,
                met,
                clo,
                model=Models.iso_7730_2005.value,
            ).pmv

            # Check for NaN and handle retries
            if np.isnan(pmv_value):
                for _retry in range(retry_attempts):
                    adjustment_factor = retry_adjustment_factor
                    to = initial_to  # Reset to initial temperature for retry
                    pmv_value = pmv_ppd_iso(
                        to,
                        to,
                        v,
                        rh,
                        met,
                        clo,
                        model=Models.iso_7730_2005.value,
                    ).pmv

                    if abs(pmv_value) < tolerance:
                        return to

                    # Adjust the temperature during retry
                    to = to - pmv_value / adjustment_factor

                # If retries fail, log a warning and stop the loop
                return to

            # Check if the PMV is within tolerance
            if abs(pmv_value) < tolerance:
                return to

            # Adjust the operative temperature using the adjustment factor
            to = to - pmv_value / adjustment_factor

        return to

    # TODO check the name of the function and the docstring
    def _reset_setpt(self) -> JOS3Output:
        """Reset set-point temperatures under steady state conditions. For a nude person in a reference environment,
        of 50% RH, 0.1 m/s air velocity, and par=1.25.

        Set-point temperatures are hypothetical core or skin temperatures in a thermally neutral state
        when at rest, similar to room set-point temperatures for air conditioning. This function is
        used during initialization to calculate the set-point temperatures as a reference for
        thermoregulation. Note that input parameters (tdb, tr, rh, v, clo, par) and body temperatures
        are also reset.

        Returns
        -------
        dict
            Parameters of the JOS-3 model.

        """
        # Set operative temperature under PMV=0 environment
        # TODO shall these be the reference values for naked?
        par: float = 1.25  # Physical activity ratio
        met = self.bmr * par / met_to_w_m2  # [met]
        rh = 50
        v = 0.1
        clo = 0
        self.to = self._calculate_operative_temp_when_pmv_is_zero(
            met=met,
            rh=rh,
            v=v,
            clo=clo,
        )
        self.rh = rh
        self.v = v
        self.clo = clo
        self.par = par  # Physical activity ratio

        # Steady-calculation
        self.options["ava_zero"] = True
        # TODO how these values range, and dtime where selected?
        for _ in range(10):
            dict_out = self._run(dtime=60000, passive=True)

        # Set new set-point temperatures for core and skin
        # TODO why are not we simply just setting these values and not returning anything?
        self.cr_set_point = self.t_core
        self.sk_set_point = self.t_skin
        self.options["ava_zero"] = False

        return dict_out

    def simulate(self, times: int, dtime=60, output: bool = True) -> None:
        """Run the JOS-3 model simulation.

        This method executes the JOS-3 model for a specified number of loops, simulating
        human thermoregulation over time. The results of each simulation step can be recorded
        and accessed later.

        Parameters
        ----------
        times : int
            Number of loops of the simulation.
        dtime : int or float, optional
            Time delta in seconds for each simulation step. Default is 60.
        output : bool, optional
            If True, records the parameters at each simulation step. Default is True.

        Returns
        -------
        None

        Examples
        --------
        Create an instance of the JOS3 class and run the simulation:

        .. code-block:: python
            :emphasize-lines: 7

            from pythermalcomfort.models import JOS3

            # Create an instance of the JOS3 class
            jos3_model = JOS3(height=1.75, weight=70, age=25, sex="female")

            # Run the simulation for 30 loops with a time delta of 60 seconds
            jos3_model.simulate(times=30, dtime=60)

            # Access the results
            results = jos3_model.dict_results()
            print(results)

        """
        # Loop through the simulation for the given number of times
        for _ in range(times):
            # Increment the elapsed time by the time delta
            self._time += dt.timedelta(0, dtime)

            # Execute the simulation step
            dict_data: JOS3Output = self._run(dtime=dtime, output=output)

            # If output is True, append the results to the history
            if output:
                self._history.append(dict_data)

    def _run(self, dtime=60, passive=False, output=True) -> JOS3Output:
        """Run a single cycle of the JOS-3 model simulation.

        This method calculates various thermoregulation parameters using the input data,
        such as convective and radiative heat transfer coefficients, operative temperature,
        heat resistance, and blood flow rates. It also computes thermogenesis by shivering
        and non-shivering, basal thermogenesis, and thermogenesis by work. The function
        constructs and solves matrices to determine the new body temperature distribution.

        Parameters
        ----------
        dtime : int or float, optional
            Time step in seconds for the simulation. Default is 60.
        passive : bool, optional
            If True, the set-point temperature for thermoregulation is set to the current
            body temperature, simulating a passive model. Default is False.
        output : bool, optional
            If True, returns a dictionary of the simulation results. Default is True.

        Returns
        -------
        JOS3Output
            A dictionary containing the simulation results, including parameters such as
            model time, mean skin temperature, skin temperature, core temperature,
            mean skin wettedness, skin wettedness, weight loss by evaporation and respiration,
            cardiac output, total thermogenesis, respiratory heat loss, and total heat loss
            from the skin to the environment.

        """
        # Compute convective and radiative heat transfer coefficient [W/(m2*K)]
        # based on posture, air velocity, air temperature, and skin temperature.
        # Manual setting is possible by setting self._hc and self._hr.
        # Compute heat and evaporative heat resistance [m2.K/W], [m2.kPa/W]

        # Get core and skin temperatures
        tcr = self.t_core
        tsk = self.t_skin

        # Convective and radiative heat transfer coefficients [W/(m2*K)]
        hc = threg.fixed_hc(
            threg.conv_coef(
                posture=self._posture,
                v=self._v,
                tdb=self._tdb,
                t_skin=tsk,
            ),
            self._v,
        )
        hr = threg.fixed_hr(
            threg.rad_coef(
                self._posture,
            ),
        )

        # Manually set convective and radiative heat transfer coefficients if necessary
        if self._hc is not None:
            hc = self._hc
        if self._hr is not None:
            hr = self._hr

        # Compute operative temp. [°C], clothing heat and evaporative resistance [m2.K/W], [m2.kPa/W]
        # Operative temp. [°C]
        to = threg.operative_temp(
            self._tdb,
            self._tr,
            hc,
            hr,
        )
        # Clothing heat resistance [m2.K/W]
        r_t = threg.dry_r(hc, hr, self._clo)
        # Clothing evaporative resistance [m2.kPa/W]
        r_et = threg.wet_r(hc, self._clo, self._iclo, lewis_rate=16.5)

        # ------------------------------------------------------------------
        # Thermoregulation
        # 1) Sweating
        # 2) Vasoconstriction, Vasodilation
        # 3) Shivering and non-shivering thermogenesis
        # ------------------------------------------------------------------

        # Compute the difference between the set-point temperature and body temperatures
        # and other thermoregulation parameters.
        # If running a passive model, the set-point temperature of thermoregulation is
        # set to the current body temperature.

        # set-point temperature for thermoregulation
        if passive:
            setpt_cr = tcr.copy()
            setpt_sk = tsk.copy()
        else:
            setpt_cr = self.cr_set_point.copy()
            setpt_sk = self.sk_set_point.copy()

        # Error signal = Difference between set-point and body temperatures
        err_cr = tcr - setpt_cr
        err_sk = tsk - setpt_sk

        # SWEATING THERMOREGULATION
        # Skin wettedness [-], e_skin, e_max, e_sweat [W]
        # Calculate skin wettedness, sweating heat loss, maximum sweating rate, and total sweat rate
        wet, e_sk, e_max, e_sweat = threg.evaporation(
            err_cr,
            err_sk,
            tsk,
            self._tdb,
            self._rh,
            r_et,
            self._height,
            self._weight,
            self._bsa_equation,
            self._age,
        )

        # VASOCONSTRICTION, VASODILATION
        # Calculate skin blood flow and basal skin blood flow [L/h]
        bf_skin = threg.skin_blood_flow(
            err_cr,
            err_sk,
            self._height,
            self._weight,
            self._bsa_equation,
            self._age,
            self._ci,
        )

        # Calculate hands and feet's AVA blood flow [L/h]
        bf_ava_hand, bf_ava_foot = threg.ava_blood_flow(
            err_cr,
            err_sk,
            self._height,
            self._weight,
            self._bsa_equation,
            self._age,
            self._ci,
        )
        if self.options["ava_zero"] and passive:
            bf_ava_hand = 0
            bf_ava_foot = 0

        # SHIVERING AND NON-SHIVERING
        # Calculate shivering thermogenesis [W]
        q_shiv = threg.shivering(
            err_cr,
            err_sk,
            tcr,
            tsk,
            self._height,
            self._weight,
            self._bsa_equation,
            self._age,
            self._sex,
            dtime,
            self.options,
        )

        # Calculate non-shivering thermogenesis (NST) [W]
        if self.options["nonshivering_thermogenesis"]:
            q_nst = threg.nonshivering(
                err_sk=err_sk,
                height=self._height,
                weight=self._weight,
                bsa_equation=self._bsa_equation,
                age=self._age,
                cold_acclimation=self.options["cold_acclimated"],
                batpositive=self.options["bat_positive"],
            )
        else:  # not consider NST
            q_nst = np.zeros(Default.num_body_parts)

        # ------------------------------------------------------------------
        # Thermogenesis
        # ------------------------------------------------------------------

        # Calculate local basal metabolic rate (BMR) [W]
        q_bmr_local = threg.local_mbase(
            self._height,
            self._weight,
            self._age,
            self._sex,
            self._bmr_equation,
        )
        # Calculate overall basal metabolic rate (BMR) [W]
        q_bmr_total = sum([m.sum() for m in q_bmr_local])

        # Calculate thermogenesis by work [W]
        q_work = threg.local_q_work(q_bmr_total, self._par)

        # Calculate the sum of thermogenesis in core, muscle, fat, skin [W]
        (
            q_thermogenesis_core,
            q_thermogenesis_muscle,
            q_thermogenesis_fat,
            q_thermogenesis_skin,
        ) = threg.sum_m(
            q_bmr_local,
            q_work,
            q_shiv,
            q_nst,
        )
        q_thermogenesis_total = (
            q_thermogenesis_core.sum()
            + q_thermogenesis_muscle.sum()
            + q_thermogenesis_fat.sum()
            + q_thermogenesis_skin.sum()
        )

        # ------------------------------------------------------------------
        # Others
        # ------------------------------------------------------------------
        # Calculate blood flow in core, muscle, fat [L/h]
        bf_core, bf_muscle, bf_fat = threg.cr_ms_fat_blood_flow(
            q_work,
            q_shiv,
            self._height,
            self._weight,
            self._bsa_equation,
            self._age,
            self._ci,
        )

        # Calculate heat loss by respiratory
        p_a = antoine(self._tdb) * self._rh / 100
        res_sh, res_lh = threg.resp_heat_loss(
            self._tdb[0],
            p_a[0],
            q_thermogenesis_total,
        )

        # Calculate sensible heat loss [W]
        shl_sk = (tsk - to) / r_t * self._bsa

        # Calculate cardiac output [L/h]
        co = threg.sum_bf(bf_core, bf_muscle, bf_fat, bf_skin, bf_ava_hand, bf_ava_foot)

        # Calculate weight loss rate by evaporation [g/sec]
        wlesk = (e_sweat + 0.06 * e_max) / 2418
        wleres = res_lh / 2418

        # ------------------------------------------------------------------
        # Matrix
        # This code section is focused on constructing and calculating
        # various matrices required for modeling the thermoregulation
        # of the human body.
        # Since JOS-3 has 85 thermal nodes, the determinant of 85*85 is to be solved.
        # ------------------------------------------------------------------

        # Matrix A = Matrix for heat exchange due to blood flow and conduction occurring between tissues
        # (85, 85,) ndarray

        # Calculate the blood flow in arteries and veins for core, muscle, fat, skin,
        # and arteriovenous anastomoses (AVA) in hands and feet,
        # and combines them into two arrays:
        # 1) bf_local for the local blood flow and 2) bf_whole for the whole-body blood flow.
        # These arrays are then combined to form arr_bf.
        bf_art, bf_vein = matrix.vessel_blood_flow(
            bf_core,
            bf_muscle,
            bf_fat,
            bf_skin,
            bf_ava_hand,
            bf_ava_foot,
        )
        bf_local = matrix.local_arr(
            bf_core,
            bf_muscle,
            bf_fat,
            bf_skin,
            bf_ava_hand,
            bf_ava_foot,
        )
        bf_whole = matrix.whole_body(bf_art, bf_vein, bf_ava_hand, bf_ava_foot)
        arr_bf = np.zeros((NUM_NODES, NUM_NODES))
        arr_bf += bf_local
        arr_bf += bf_whole

        # Adjusts the units of arr_bf from [W/K] to [/sec] and then to [-]
        # by dividing by the heat capacity self._cap and multiplying by the time step dtime.
        arr_bf /= self._cap.reshape((NUM_NODES, 1))  # Change unit [W/K] to [/sec]
        arr_bf *= dtime  # Change unit [/sec] to [-]

        # Performs similar unit conversions for the convective heat transfer coefficient array arr_cdt
        # (also divided by self._cap and multiplied by dtime).
        arr_cdt = self._cdt.copy()
        arr_cdt /= self._cap.reshape((NUM_NODES, 1))  # Change unit [W/K] to [/sec]
        arr_cdt *= dtime  # Change unit [/sec] to [-]

        # Matrix B = Matrix for heat transfer between skin and environment
        arr_b = np.zeros(NUM_NODES)
        arr_b[INDEX["skin"]] += 1 / r_t * self._bsa
        arr_b /= self._cap  # Change unit [W/K] to [/sec]
        arr_b *= dtime  # Change unit [/sec] to [-]

        # Calculate the off-diagonal and diagonal elements of the matrix A,
        # which represents the heat transfer coefficients between different parts of the body,
        # and combines them to form the full matrix A (arrA).
        # Then, the inverse of matrix A is computed (arrA_inv).
        arr_a_tria = -(arr_cdt + arr_bf)

        arr_a_dia = arr_cdt + arr_bf
        arr_a_dia = arr_a_dia.sum(axis=1) + arr_b
        arr_a_dia = np.diag(arr_a_dia)
        arr_a_dia += np.eye(NUM_NODES)

        arr_a = arr_a_tria + arr_a_dia
        arr_a_inv = np.linalg.inv(arr_a)

        # Matrix Q = Matrix for heat generation rate from thermogenesis, respiratory, sweating,
        # and extra heat gain processes in different body parts.

        # Matrix Q [W] / [J/K] * [sec] = [-]
        # Thermogensis
        arr_q = np.zeros(NUM_NODES)
        arr_q[INDEX["core"]] += q_thermogenesis_core
        arr_q[INDEX["muscle"]] += q_thermogenesis_muscle[VINDEX["muscle"]]
        arr_q[INDEX["fat"]] += q_thermogenesis_fat[VINDEX["fat"]]
        arr_q[INDEX["skin"]] += q_thermogenesis_skin

        # Respiratory [W]
        arr_q[INDEX["core"][2]] -= res_sh + res_lh  # chest core

        # Sweating [W]
        arr_q[INDEX["skin"]] -= e_sk

        # Extra heat gain [W]
        arr_q += self.ex_q.copy()

        arr_q /= self._cap  # Change unit [W]/[J/K] to [K/sec]
        arr_q *= dtime  # Change unit [K/sec] to [K]

        # Boundary batrix [°C]
        arr_to = np.zeros(NUM_NODES)
        arr_to[INDEX["skin"]] += to

        # Combines the current body temperature, the boundary matrix, and the heat generation matrix
        # to calculate the new body temperature distribution (arr).
        arr = self._t_body + arr_b * arr_to + arr_q

        # ------------------------------------------------------------------
        # New body temp. [°C]
        # ------------------------------------------------------------------
        self._t_body = np.dot(arr_a_inv, arr)

        # ------------------------------------------------------------------
        # Output parameters
        # ------------------------------------------------------------------

        return JOS3Output(
            simulation_time=self._time,
            dt=dtime,
            t_skin_mean=np.round(self.t_skin_mean, 2),
            t_skin=pass_values_to_jos3_body_parts(self.t_skin),
            t_core=pass_values_to_jos3_body_parts(self.t_core),
            w_mean=round(np.average(wet, weights=Default.local_bsa), 2),
            w=pass_values_to_jos3_body_parts(wet),
            weight_loss_by_evap_and_res=round(wlesk.sum() + wleres, 5),
            cardiac_output=round(co, 1),
            q_thermogenesis_total=round(q_thermogenesis_total, 2),
            q_res=round(res_sh + res_lh, 2),
            q_skin2env=pass_values_to_jos3_body_parts(shl_sk + e_sk),
            height=self._height,
            weight=self._weight,
            bsa=pass_values_to_jos3_body_parts(self._bsa),
            fat=self._fat,
            sex=self._sex,
            age=self._age,
            t_core_set=pass_values_to_jos3_body_parts(setpt_cr),
            t_skin_set=pass_values_to_jos3_body_parts(setpt_sk),
            t_cb=round(self.t_cb, 2),
            t_artery=pass_values_to_jos3_body_parts(self.t_artery),
            t_vein=pass_values_to_jos3_body_parts(self.t_vein),
            t_superficial_vein=pass_values_to_jos3_body_parts(self.t_superficial_vein),
            t_muscle=pass_values_to_jos3_body_parts(
                self.t_muscle,
                body_parts=["head", "pelvis"],
            ),
            t_fat=pass_values_to_jos3_body_parts(
                self.t_fat,
                body_parts=["head", "pelvis"],
            ),
            to=pass_values_to_jos3_body_parts(to),
            r_t=pass_values_to_jos3_body_parts(r_t, 3),
            r_et=pass_values_to_jos3_body_parts(r_et, 3),
            tdb=pass_values_to_jos3_body_parts(self._tdb),
            tr=pass_values_to_jos3_body_parts(self._tr),
            rh=pass_values_to_jos3_body_parts(self._rh),
            v=pass_values_to_jos3_body_parts(self._v),
            par=self._par,
            clo=pass_values_to_jos3_body_parts(self._clo),
            e_skin=pass_values_to_jos3_body_parts(e_sk),
            e_max=pass_values_to_jos3_body_parts(e_max),
            e_sweat=pass_values_to_jos3_body_parts(e_sweat),
            bf_core=pass_values_to_jos3_body_parts(bf_core),
            bf_muscle=pass_values_to_jos3_body_parts(
                bf_muscle[VINDEX["muscle"]],
                body_parts=["head", "pelvis"],
            ),
            bf_fat=pass_values_to_jos3_body_parts(
                bf_fat[VINDEX["fat"]],
                body_parts=["head", "pelvis"],
            ),
            bf_skin=pass_values_to_jos3_body_parts(bf_skin),
            bf_ava_hand=np.round(bf_ava_hand, 2),
            bf_ava_foot=np.round(bf_ava_foot, 2),
            q_bmr_core=pass_values_to_jos3_body_parts(q_bmr_local[0]),
            q_bmr_muscle=pass_values_to_jos3_body_parts(
                q_bmr_local[1][VINDEX["muscle"]],
                body_parts=["head", "pelvis"],
            ),
            q_bmr_fat=pass_values_to_jos3_body_parts(
                q_bmr_local[2][VINDEX["fat"]],
                body_parts=["head", "pelvis"],
            ),
            q_bmr_skin=pass_values_to_jos3_body_parts(q_bmr_local[3]),
            q_work=pass_values_to_jos3_body_parts(q_work),
            q_shiv=pass_values_to_jos3_body_parts(q_shiv),
            q_nst=pass_values_to_jos3_body_parts(q_nst),
            q_thermogenesis_core=pass_values_to_jos3_body_parts(q_thermogenesis_core),
            q_thermogenesis_muscle=pass_values_to_jos3_body_parts(
                q_thermogenesis_muscle[VINDEX["muscle"]],
                body_parts=["head", "pelvis"],
            ),
            q_thermogenesis_fat=pass_values_to_jos3_body_parts(
                q_thermogenesis_fat[VINDEX["fat"]],
                body_parts=["head", "pelvis"],
            ),
            q_thermogenesis_skin=pass_values_to_jos3_body_parts(q_thermogenesis_skin),
            q_skin2env_sensible=pass_values_to_jos3_body_parts(shl_sk),
            q_skin2env_latent=pass_values_to_jos3_body_parts(e_sk),
            q_res_sensible=np.round(res_sh, 2),
            q_res_latent=np.round(res_lh, 2),
        )

    def results(self) -> JOS3Output:
        """Consolidate the results into a single JOS3Output instance. This makes
        it very easy to access the time series data for each parameter.

        Returns
        -------
        JOS3Output
            A single JOS3Output instance where each attribute is an array of values
            representing the time series data for that attribute.

        Examples
        --------
        .. code-block:: python

            from pythermalcomfort.models import JOS3

            model = JOS3(height=1.75, weight=70, age=25, sex="female")
            model.simulate(times=3, dtime=60)
            output = model.results()
            print(output.t_skin_mean)
            print(output.t_skin.head)

        """
        merged_data = defaultdict(list)

        for output in self._history:
            for field in fields(JOS3Output):
                value = getattr(output, field.name)
                if isinstance(value, JOS3BodyParts):
                    if field.name not in merged_data:
                        merged_data[field.name] = defaultdict(list)
                    for part in fields(JOS3BodyParts):
                        merged_data[field.name][part.name].append(
                            getattr(value, part.name),
                        )
                else:
                    merged_data[field.name].append(value)

        for key, value in merged_data.items():
            if isinstance(value, defaultdict):
                merged_data[key] = JOS3BodyParts(
                    **{k: np.array(v) for k, v in value.items()},
                )
            else:
                merged_data[key] = np.array(value)

        return JOS3Output(**merged_data)

    def dict_results(self):
        """Get simulation results as a dictionary.

        This method returns the results of the JOS-3 model simulation as a dictionary,
        where each key corresponds to a specific parameter and the value is an array
        containing the time series data for that parameter.

        Returns
        -------
        dict
            A dictionary of the simulation results. Keys are parameter names, and values
            are arrays containing the time series data for each parameter.

        Examples
        --------
        Access the results after running a simulation:

        .. code-block:: python
            :emphasize-lines: 3

            jos3_model = JOS3(height=1.75, weight=70, age=25, sex="female")
            jos3_model.simulate(times=30, dtime=60)
            results = jos3_model.dict_results()
            print(
                results["t_skin_mean"]
            )  # Access the mean skin temperature time series

        """
        if not self._history:
            print("The model has no data.")
            return None

        def check_word_contain(word, *args):
            """Check if word contains *args."""
            bool_filter = False
            for arg in args:
                if arg in word:
                    bool_filter = True
            return bool_filter

        # Set column titles
        # If the values are iter, add the body names as suffix words.
        # If the values are not iter and the single value data, convert it to iter.
        key2keys = {}  # Column keys
        for key, value in self._history[0].__dict__.items():
            try:
                if isinstance(value, JOS3BodyParts):
                    length = len(value.__dict__)
                else:
                    length = len(value)
                if isinstance(value, str):
                    keys = [key]  # str is iter. Convert to list without suffix
                elif check_word_contain(key, "sve", "sfv", "superficialvein"):
                    keys = [
                        key + "_" + JOS3BodyParts.get_attribute_names()[i]
                        for i in VINDEX["sfvein"]
                    ]
                elif check_word_contain(key, "ms", "muscle"):
                    keys = [
                        key + "_" + JOS3BodyParts.get_attribute_names()[i]
                        for i in VINDEX["muscle"]
                    ]
                elif check_word_contain(key, "fat"):
                    keys = [
                        key + "_" + JOS3BodyParts.get_attribute_names()[i]
                        for i in VINDEX["fat"]
                    ]
                elif length == Default.num_body_parts:  # if data contains 17 values
                    keys = [
                        key + "_" + bn for bn in JOS3BodyParts.get_attribute_names()
                    ]
                else:
                    keys = [
                        key + "_" + JOS3BodyParts.get_attribute_names()[i]
                        for i in range(length)
                    ]
            except TypeError:  # if the value is not iter.
                keys = [key]  # convert to iter
            key2keys.update({key: keys})

        data = []
        for _i, dictout in enumerate(self._history):
            row = {}
            for key, value in dictout.__dict__.items():
                keys = key2keys[key]
                # make list if value is not iter
                values = [value] if len(keys) == 1 else value.__dict__
                row.update(dict(zip(keys, values, strict=False)))
            data.append(row)

        out_dict = dict(
            zip(data[0].keys(), [[] for _ in range(len(data[0].keys()))], strict=False)
        )
        for row in data:
            for k in data[0].keys():
                out_dict[k].append(row[k])
        return out_dict

    def to_csv(
        self,
        path: str = None,
        folder: str = None,
        unit: bool = True,
        meaning: bool = True,
    ) -> None:
        """Export results as csv format.

        Parameters
        ----------
        path : str, optional
            Output path. If you don't use the default file name, set a name.
            The default is None.
        folder : str, optional
            Output folder. If you use the default file name with the current time,
            set a only folder path.
            The default is None.
        unit : bool, optional
            Write units in csv file. The default is True.
        meaning : bool, optional
            Write meanings of the parameters in csv file. The default is True.

        Returns
        -------
        None

        Examples
        --------

        .. code-block:: python

            from pythermalcomfort.models import JOS3

            # Create an instance of the JOS3 model
            model = JOS3()

            # Simulate for 60 steps
            model.simulate(60)

            # Export results to a CSV file
            model.to_csv()

        """
        # Use the model name and current time as default output file name
        if path is None:
            now_time = dt.datetime.now().strftime("%Y%m%d-%H%M%S")
            path = f"{now_time}.csv"
            if folder:
                os.makedirs(folder, exist_ok=True)
                path = folder + os.sep + path
        elif not ((path[-4:] == ".csv") or (path[-4:] == ".txt")):
            path += ".csv"

        # Get simulation results as a dictionary
        dict_out = self.dict_results()

        # Get column names, units and meanings
        columns = [k for k in dict_out.keys()]
        units = []
        meanings = []
        for col in columns:
            param, body_name = remove_body_name(col)
            if param in ALL_OUT_PARAMS:
                u = ALL_OUT_PARAMS[param]["unit"]
                units.append(u)

                m = ALL_OUT_PARAMS[param]["meaning"]
                if body_name:
                    # Replace underscores with spaces
                    body_name_with_spaces = body_name.replace("_", " ")
                    meanings.append(m.replace("each body part", body_name_with_spaces))
                else:
                    meanings.append(m)
            else:
                units.append("")
                meanings.append("")

        # Write to csv file
        with open(path, "w", newline="", encoding="utf-8-sig") as f:
            writer = csv.writer(f)
            writer.writerow(list(columns))
            if unit:
                writer.writerow(units)
            if meaning:
                writer.writerow(meanings)
            for i in range(len(dict_out["cycle_time"])):
                row = []
                for k in columns:
                    row.append(dict_out[k][i])
                writer.writerow(row)

    def _set_ex_q(self, tissue, value):
        """Set extra heat gain by tissue name.

        Parameters
        ----------
        tissue : str
            Tissue name. "core", "skin", or "artery".... If you set value to
            head muscle and other segment's core, set "all_muscle".
        value : int, float, array
            Heat gain [W]

        Returns
        -------
        array
            Extra heat gain of model.

        """
        self.ex_q[INDEX[tissue]] = value
        return self.ex_q

    @property
    def tdb(self):
        """Dry-bulb air temperature. The setter accepts int, float, dict, list, ndarray.
        The inputs are used to create a 17-element array. dict should be passed with
        BODY_NAMES as keys.

        Returns
        -------
        ndarray
            A NumPy array of shape (17).

        """
        return self._tdb

    @tdb.setter
    def tdb(self, inp):
        self._tdb = to_array_body_parts(inp)

    @property
    def tr(self):
        """Tr : numpy.ndarray (17) Mean radiant temperature [°C]."""
        return self._tr

    @tr.setter
    def tr(self, inp):
        self._tr = to_array_body_parts(inp)

    @property
    def to(self):
        """To : numpy.ndarray (17) Operative temperature [°C]."""
        hc = threg.fixed_hc(
            threg.conv_coef(
                self._posture,
                self._v,
                self._tdb,
                self.t_skin,
            ),
            self._v,
        )
        hr = threg.fixed_hr(
            threg.rad_coef(
                self._posture,
            ),
        )
        return threg.operative_temp(
            self._tdb,
            self._tr,
            hc,
            hr,
        )

    @to.setter
    # TODO to should not be a setter otherwise people can erroneously overrite it
    def to(self, inp):
        self._tdb = to_array_body_parts(inp)
        self._tr = to_array_body_parts(inp)

    @property
    def rh(self):
        """Rh : numpy.ndarray (17) Relative humidity [%]."""
        return self._rh

    @rh.setter
    def rh(self, inp):
        self._rh = to_array_body_parts(inp)

    @property
    def v(self):
        """V : numpy.ndarray (17) Air velocity [m/s]."""
        return self._v

    @v.setter
    def v(self, inp):
        self._v = to_array_body_parts(inp)

    @property
    def posture(self):
        """Posture : str Current JOS3 posture."""
        return self._posture

    @posture.setter
    def posture(self, inp):
        if inp == 0:
            self._posture = Postures.standing.value
        elif inp == 1:
            self._posture = Postures.sitting.value
        elif inp == 2:
            self._posture = Postures.lying.value
        elif isinstance(inp, str):
            if inp.lower() == Postures.standing.value:
                self._posture = Postures.standing.value
            elif inp.lower() in [Postures.sitting.value, Postures.sedentary.value]:
                self._posture = Postures.sitting.value
            elif inp.lower() in [Postures.lying.value, Postures.supine.value]:
                self._posture = Postures.lying.value
        else:
            self._posture = Postures.standing.value
            print(
                f"posture must be 0={Postures.standing.value}, 1={Postures.sitting.value} or 2={Postures.lying.value}.",
            )
            print(f"posture was set {Postures.standing.value}.")

    @property
    def clo(self):
        """Clo : numpy.ndarray (17) Clothing insulation [clo]."""
        return self._clo

    @clo.setter
    def clo(self, inp):
        self._clo = to_array_body_parts(inp)

    @property
    def par(self):
        """Par : float Physical activity ratio [-].This equals the ratio of metabolic rate to basal metabolic rate. par of sitting quietly is 1.2."""
        return self._par

    @par.setter
    def par(self, inp):
        self._par = inp

    @property
    def t_body(self):
        """t_body : numpy.ndarray (85,) All segment temperatures of JOS-3."""
        return self._t_body

    @t_body.setter
    def t_body(self, inp):
        self._t_body = inp.copy()

    @property
    def bsa(self):
        """Bsa : numpy.ndarray (17) Body surface areas by local body segments [m2]."""
        return self._bsa.copy()

    @property
    def r_t(self):
        """r_t : numpy.ndarray (17) Dry heat resistances between the skin and ambience areas by local body segments [(m2*K)/W]."""
        hc = threg.fixed_hc(
            threg.conv_coef(
                self._posture,
                self._v,
                self._tdb,
                self.t_skin,
            ),
            self._v,
        )
        hr = threg.fixed_hr(
            threg.rad_coef(
                self._posture,
            ),
        )
        return threg.dry_r(hc, hr, self._clo)

    @property
    def r_et(self):
        """r_et : numpy.ndarray (17) w (Evaporative) heat resistances between the skin and
        ambience areas by local body segments [(m2*kPa)/W].
        """
        hc = threg.fixed_hc(
            threg.conv_coef(
                self._posture,
                self._v,
                self._tdb,
                self.t_skin,
            ),
            self._v,
        )
        return threg.wet_r(hc=hc, clo=self._clo, i_clo=self._iclo, lewis_rate=16.5)

    @property
    def w(self):
        """W : numpy.ndarray (17) Skin wettedness on local body segments [-]."""
        err_cr = self.t_core - self.cr_set_point
        err_sk = self.t_skin - self.sk_set_point
        wet, *_ = threg.evaporation(
            err_cr=err_cr,
            err_sk=err_sk,
            t_skin=self.t_skin,
            tdb=self._tdb,
            rh=self._rh,
            ret=self.r_et,
            height=self._height,
            weight=self._weight,
            bsa_equation=self._bsa_equation,
            age=self._age,
        )
        return wet

    @property
    def w_mean(self):
        """w_mean : float Mean skin wettedness of the whole body [-]."""
        wet = self.w
        return np.average(wet, weights=Default.local_bsa)

    @property
    def t_skin_mean(self) -> float:
        """t_skin_mean : float Mean skin temperature of the whole body [°C]."""
        return float(np.average(self._t_body[INDEX["skin"]], weights=Default.local_bsa))

    # TODO all the properties should be returning JOS3BodyParts
    @property
    def t_skin(self) -> np.ndarray[float]:
        """t_skin : numpy.ndarray (17) Skin temperatures by the local body segments [°C]."""
        return self._t_body[INDEX["skin"]].copy()

    @t_skin.setter
    def t_skin(self, inp):
        self._t_body[INDEX["skin"]] = to_array_body_parts(inp)

    @property
    def t_core(self) -> np.ndarray[float]:
        """t_core : numpy.ndarray (17) Skin temperatures by the local body segments [°C]."""
        return self._t_body[INDEX["core"]].copy()

    @property
    def t_cb(self) -> float:
        """t_cb : float Temperature at central blood pool [°C]."""
        return float(self._t_body[0].copy())

    @property
    def t_artery(self) -> np.ndarray[float]:
        """t_artery : numpy.ndarray (17) Arterial temperatures by the local body segments [°C]."""
        return self._t_body[INDEX["artery"]].copy()

    @property
    def t_vein(self) -> np.ndarray[float]:
        """t_vein : numpy.ndarray (17) Vein temperatures by the local body segments [°C]."""
        return self._t_body[INDEX["vein"]].copy()

    @property
    def t_superficial_vein(self) -> np.ndarray[float]:
        """t_superficial_vein : numpy.ndarray (12,) Superficial vein temperatures by the local body segments [°C]."""
        return self._t_body[INDEX["sfvein"]].copy()

    @property
    def t_muscle(self) -> np.ndarray[float]:
        """t_muscle : numpy.ndarray (2,) Muscle temperatures of head and pelvis [°C]."""
        return self._t_body[INDEX["muscle"]].copy()

    @property
    def t_fat(self) -> np.ndarray[float]:
        """t_fat : numpy.ndarray (2,) fat temperatures of head and pelvis  [°C]."""
        return self._t_body[INDEX["fat"]].copy()

    @property
    def body_names(self) -> np.ndarray[float]:
        """body_names : list JOS3 body names."""
        return JOS3BodyParts.get_attribute_names()

    @property
    def bmr(self) -> float:
        """Bmr : float Basal metabolic rate [W/m2]."""
        tcr = threg.basal_met(
            self._height,
            self._weight,
            self._age,
            self._sex,
            self._bmr_equation,
        )
        return tcr / self.bsa.sum()



from pythermalcomfort.models import net
from tests.conftest import Urls, retrieve_reference_table, validate_result


def test_net(get_test_url, retrieve_data) -> None:
    """Test the NET model with various inputs."""
    reference_table = retrieve_reference_table(
        get_test_url,
        retrieve_data,
        Urls.NET.name,
    )
    tolerance = reference_table["tolerance"]

    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        outputs = entry["outputs"]
        result = net(**inputs)

        validate_result(result, outputs, tolerance)



import math

import numpy as np

from pythermalcomfort.models._pmv_ppd_optimized import _pmv_ppd_optimized


class TestPmvPpdOptimized:
    """Test cases for the PMV and PPD optimized model."""

    #  Returns NaN for invalid input values

    def test_pmv_ppd_optimized(self) -> None:
        """Test the optimized PMV and PPD function with various inputs."""
        assert math.isclose(
            _pmv_ppd_optimized(25, 25, 0.3, 50, 1.5, 0.7, 0),
            0.55,
            abs_tol=0.01,
        )

        np.testing.assert_equal(
            np.around(_pmv_ppd_optimized([25, 25], 25, 0.3, 50, 1.5, 0.7, 0), 2),
            [0.55, 0.55],
        )

    #  The function returns the correct PMV value for typical input values.
    def test_pmv_typical_input(self) -> None:
        """Test the optimized PMV and PPD function with typical input values."""
        # Typical input values
        tdb = 25
        tr = 23
        vr = 0.1
        rh = 50
        met = 1.2
        clo = 0.5
        wme = 0

        expected_pmv = -0.197

        assert math.isclose(
            _pmv_ppd_optimized(tdb, tr, vr, rh, met, clo, wme),
            expected_pmv,
            abs_tol=0.01,
        )

    #  The function returns the correct PMV value for extreme input values.
    def test_pmv_extreme_input(self) -> None:
        """Test the optimized PMV and PPD function with extreme input values."""
        # Extreme input values
        tdb = 35
        tr = 45
        vr = 2
        rh = 10
        met = 2.5
        clo = 1.5
        wme = 1

        expected_pmv = 1.86

        assert math.isclose(
            _pmv_ppd_optimized(tdb, tr, vr, rh, met, clo, wme),
            expected_pmv,
            abs_tol=0.01,
        )

    #  The function returns NaN if any of the input values are NaN.
    def test_nan_input_values(self) -> None:
        """Test the optimized PMV and PPD function with NaN input values."""
        tdb = float("nan")
        tr = 23
        vr = 0.1
        rh = 50
        met = 1.2
        clo = 0.5
        wme = 0

        assert math.isnan(_pmv_ppd_optimized(tdb, tr, vr, rh, met, clo, wme))

    #  The function returns NaN if any of the input values are infinite.
    def test_infinite_input_values(self) -> None:
        """Test the optimized PMV and PPD function with infinite input values."""
        # Input values with infinity
        tdb = float("inf")
        tr = 23
        vr = 0.1
        rh = 50
        met = 1.2
        clo = 0.5
        wme = 0

        assert math.isnan(_pmv_ppd_optimized(tdb, tr, vr, rh, met, clo, wme))



from __future__ import annotations

import numpy as np

from pythermalcomfort.classes_input import ATInputs
from pythermalcomfort.classes_return import AT
from pythermalcomfort.utilities import psy_ta_rh


def at(
    tdb: float | list[float],
    rh: float | list[float],
    v: float | list[float],
    q: float | list[float] = None,
    round_output: bool = True,
) -> AT:
    """Calculate the Apparent Temperature (AT). The AT is defined as the
    temperature at the reference humidity level producing the same amount of
    discomfort as that experienced under the current ambient temperature,
    humidity, and solar radiation [Steadman1984]_. In other words, the AT is an
    adjustment to the dry bulb temperature based on the relative humidity
    value. Absolute humidity with a dew point of 14°C is chosen as a reference.

    It includes the chilling effect of the wind at lower temperatures. [Blazejczyk2012]_

    .. note::
        Two formulas for AT are in use by the Australian Bureau of Meteorology: one includes
        solar radiation and the other one does not (http://www.bom.gov.au/info/thermal_stress/
        , 29 Sep 2021). Please specify q if you want to estimate AT with solar load.

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, [°C]
    rh : float or list of floats
        Relative humidity, [%]
    v : float or list of floats
        Wind speed 10m above ground level, [m/s]
    q : float or list of floats, optional
        Net radiation absorbed per unit area of body surface [W/m2]
    round_output : bool, default True
        If True, rounds the output value; if False, does not round it.

    Returns
    -------
    AT
        Dataclass containing the apparent temperature, [°C]. See :py:class:`~pythermalcomfort.classes_return.AT` for more details.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import at

        at(tdb=25, rh=30, v=0.1)
        # AT(at=24.1)

    """
    # Validate inputs
    ATInputs(tdb=tdb, rh=rh, v=v, q=q, round_output=round_output)

    # Convert lists to numpy arrays if necessary
    tdb = np.array(tdb)
    rh = np.array(rh)
    v = np.array(v)

    # Calculate vapor pressure
    p_vap = psy_ta_rh(tdb, rh).p_vap / 100

    # Calculate apparent temperature
    if q is not None:
        q = np.array(q)
        t_at = tdb + 0.348 * p_vap - 0.7 * v + 0.7 * q / (v + 10) - 4.25
    else:
        t_at = tdb + 0.33 * p_vap - 0.7 * v - 4.00

    if round_output:
        t_at = np.around(t_at, 1)

    return AT(at=t_at)



import numpy as np
import pytest

from pythermalcomfort.classes_return import WorkCapacity
from pythermalcomfort.models import work_capacity_dunne


def test_heavy_scalar_at_25() -> None:
    """Verify that a scalar WBGT of 25 with heavy intensity yields 100% capacity."""
    wc = work_capacity_dunne(25, "heavy")
    assert isinstance(wc, WorkCapacity)
    assert wc.capacity == pytest.approx(100)


def test_heavy_array_and_clipping() -> None:
    """Verify heavy intensity for an array of WBGT values and proper clipping at [0, 100]."""
    wbgt = [20, 25, 28, 30, 35]
    wc = work_capacity_dunne(wbgt, "heavy")
    expected_base = np.clip(
        100 - 25 * np.maximum(0, np.array(wbgt) - 25) ** (2 / 3),
        0,
        100,
    )
    assert isinstance(wc, WorkCapacity)
    assert np.allclose(wc.capacity, expected_base)


def test_moderate_and_light_intensity() -> None:
    """Verify moderate and light intensities scale the base capacity correctly."""
    wbgt = 30
    base = np.clip(100 - 25 * (5) ** (2 / 3), 0, 100)
    wc_med = work_capacity_dunne(wbgt, "moderate")
    wc_light = work_capacity_dunne(wbgt, "light")
    assert wc_med.capacity == pytest.approx(np.clip(base * 2, 0, 100))
    assert wc_light.capacity == pytest.approx(np.clip(base * 4, 0, 100))


def test_lower_bound_clipping() -> None:
    """Verify that capacity is clipped to 0 when WBGT is extremely high."""
    wc = work_capacity_dunne(100, "heavy")
    assert wc.capacity == pytest.approx(0)


def test_upper_bound_clipping() -> None:
    """Verify that capacity is clipped to 100 when WBGT is low and intensity is light."""
    wc = work_capacity_dunne(20, "light")
    assert wc.capacity == pytest.approx(100)


def test_invalid_intensity_raises() -> None:
    """Verify that providing an invalid intensity string raises a ValueError."""
    with pytest.raises(ValueError):
        work_capacity_dunne(25, "invalid")



from __future__ import annotations

from dataclasses import asdict

import numpy as np

from pythermalcomfort.classes_input import UseFansHeatwavesInputs
from pythermalcomfort.classes_return import UseFansHeatwaves
from pythermalcomfort.models.two_nodes_gagge import two_nodes_gagge
from pythermalcomfort.utilities import (
    Postures,
    _check_standard_compliance_array,
)


def use_fans_heatwaves(
    tdb: float | list[float],
    tr: float | list[float],
    v: float | list[float],
    rh: float | list[float],
    met: float | list[float],
    clo: float | list[float],
    wme: float | list[float] = 0,
    body_surface_area: float | list[float] = 1.8258,
    p_atm: float | list[float] = 101325,
    position: str = Postures.standing.value,
    max_skin_blood_flow: float = 80,
    max_sweating: float = 500,
    limit_inputs: bool = True,
    round_output: bool = True,
) -> UseFansHeatwaves:
    """Estimate if the conditions you have selected would cause
    heat strain.

    This occurs when either the following variables reaches its
    maximum value:

    * m_rsw Rate at which regulatory sweat is generated, [mL/h/m2]
    * w : Skin wettedness, adimensional. Ranges from 0 and 1.
    * m_bl : Skin blood flow [kg/h/m2]

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, [°C].
    tr : float or list of floats
        Mean radiant temperature, [°C].
    v : float or list of floats
        Air speed, [m/s].
    rh : float or list of floats
        Relative humidity, [%].
    met : float or list of floats
        Metabolic rate, [met].
    clo : float or list of floats
        Clothing insulation, [clo].
    wme : float or list of floats, optional
        External work, [met]. Defaults to 0.
    body_surface_area : float or list of floats, optional
        Body surface area, default value 1.8258 [m2]. Defaults to 1.8258.

        .. note::
            The body surface area can be calculated using the function
            :py:meth:`pythermalcomfort.utilities.body_surface_area`.
    p_atm : float or list of floats, optional
        Atmospheric pressure, default value 101325 [Pa]. Defaults to 101325.
    position : str, optional
        Select either "sitting" or "standing". Defaults to "standing".
    max_skin_blood_flow : float
        Maximum blood flow from the core to the skin. Defaults to 80.
    limit_inputs : bool, optional
        By default, if the inputs are outside the standard applicability limits the
        function returns nan. If False, returns values even if input values are
        outside the applicability limits of the model. Defaults to True. The
        applicability limits are 20 < tdb [°C] < 50, 20 < tr [°C] < 50,
        0.1 < v [m/s] < 4.5, 0.7 < met [met] < 2, and 0 < clo [clo] < 1.
    round_output : bool, optional
        If True, rounds output value. If False, it does not round it. Defaults to True.
    max_sweating : float or list of floats, optional
        Maximum rate at which regulatory sweat is generated, [kg/h/m2]. Defaults to 500.

    Returns
    -------
    UseFansHeatwaves
        A dataclass containing the results of using fans during heatwaves.
        See :py:class:`~pythermalcomfort.classes_return.UseFansHeatwaves` for more details.
        To access the results, use the corresponding attributes of the returned `UseFansHeatwaves` instance, e.g., `result.e_skin`.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import use_fans_heatwaves

        result = use_fans_heatwaves(tdb=35, tr=35, v=1.0, rh=50, met=1.2, clo=0.5)
        print(result.e_skin)  # 63.0

    """
    # Validate inputs using the UseFansHeatwavesInputs class
    UseFansHeatwavesInputs(
        tdb=tdb,
        tr=tr,
        v=v,
        rh=rh,
        met=met,
        clo=clo,
        wme=wme,
        body_surface_area=body_surface_area,
        p_atm=p_atm,
        position=position,
        max_skin_blood_flow=max_skin_blood_flow,
        limit_inputs=limit_inputs,
    )

    tdb = np.array(tdb)
    tr = np.array(tr)
    v = np.array(v)
    rh = np.array(rh)
    met = np.array(met)
    clo = np.array(clo)
    wme = np.array(wme)

    output = two_nodes_gagge(
        tdb,
        tr,
        v,
        rh,
        met,
        clo,
        wme=wme,
        body_surface_area=body_surface_area,
        p_atm=p_atm,
        position=position,
        max_skin_blood_flow=max_skin_blood_flow,
        round_output=False,
        max_sweating=max_sweating,
    )

    output = asdict(output)

    output_vars = [
        "e_skin",
        "e_rsw",
        "e_max",
        "q_sensible",
        "q_skin",
        "q_res",
        "t_core",
        "t_skin",
        "m_bl",
        "m_rsw",
        "w",
        "w_max",
        "heat_strain_blood_flow",
        "heat_strain_w",
        "heat_strain_sweating",
        "heat_strain",
    ]

    output["heat_strain_blood_flow"] = np.where(
        output["m_bl"] == max_skin_blood_flow,
        True,
        False,
    )
    output["heat_strain_w"] = np.where(output["w"] == output["w_max"], True, False)
    output["heat_strain_sweating"] = np.where(
        output["m_rsw"] == max_sweating,
        True,
        False,
    )

    output["heat_strain"] = np.any(
        [
            output["heat_strain_blood_flow"],
            output["heat_strain_w"],
            output["heat_strain_sweating"],
        ],
        axis=0,
    )

    output = {key: output[key] for key in output_vars}

    if limit_inputs:
        (
            tdb_valid,
            tr_valid,
            v_valid,
            rh_valid,
            met_valid,
            clo_valid,
        ) = _check_standard_compliance_array(
            standard="fan_heatwaves",
            tdb=tdb,
            tr=tr,
            v=v,
            rh=rh,
            met=met,
            clo=clo,
        )
        all_valid = ~(
            np.isnan(tdb_valid)
            | np.isnan(tr_valid)
            | np.isnan(v_valid)
            | np.isnan(met_valid)
            | np.isnan(clo_valid)
        )
        output = {key: np.where(all_valid, output[key], np.nan) for key in output_vars}

    if round_output:
        output = {key: np.around(output[key], 1) for key in output_vars}

    return UseFansHeatwaves(**output)



from __future__ import annotations

import re
from pathlib import Path

from setuptools import find_packages, setup


def read(*names: str, encoding: str = "utf8") -> str:
    """Read and return the contents of a text file located by path segments
    relative to this file's directory.

    Parameters
    ----------
    *names : str
        Sequence of path components under the project root.
    encoding : str, default "utf8"
        File encoding to use when reading.

    Returns
    -------
    str
        The full text content of the file.

    """
    base_dir = Path(__file__).parent
    file_path = base_dir.joinpath(*names)
    return file_path.read_text(encoding=encoding)


setup(
    name="pythermalcomfort",
    version="3.4.4",
    license="MIT",
    description=(
        "pythermalcomfort is a comprehensive toolkit for calculating "
        "thermal comfort indices, heat/cold stress metrics, and thermophysiological responses. "
        "It supports multiple models, including PMV, PPD, adaptive comfort, SET, "
        "UTCI, Heat Index, Wind Chill Index, and Humidex. "
        "The package also includes thermophysiological models like the two-node (Gagge) and multinode (JOS-3) models "
        "to estimate physiological responses such as core temperature, skin temperature, and skin wettedness. "
    ),
    long_description="{}\n{}".format(
        re.compile("^.. start-badges.*^.. end-badges", re.MULTILINE | re.DOTALL).sub(
            "",
            read("README.rst"),
        ),
        re.sub(":[a-z]+:`~?(.*?)`", r"``\1``", read("CHANGELOG.rst")),
    ),
    long_description_content_type="text/x-rst",
    author="Federico Tartarini",
    author_email="cbecomforttool@gmail.com",
    url="https://github.com/CenterForTheBuiltEnvironment/pythermalcomfort",
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: Education",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Education",
        "Topic :: Scientific/Engineering",
        "Topic :: Scientific/Engineering :: Atmospheric Science",
        "Topic :: Utilities",
    ],
    project_urls={
        "Documentation": "https://pythermalcomfort.readthedocs.io/",
        "Changelog": "https://pythermalcomfort.readthedocs.io/en/latest/changelog.html",
        "Issue Tracker": (
            "https://github.com/CenterForTheBuiltEnvironment/pythermalcomfort/issues"
        ),
    },
    keywords=[
        "thermal comfort",
        "pmv",
        "heat stress",
        "cold stress",
        "thermal sensation",
        "thermal physiology",
        "meteorology",
        "climate analysis",
        "discomfort",
        "comfort",
        "thermal environment",
        "built environment",
    ],
    python_requires=">=3.10.0",
    install_requires=[
        "scipy",
        "numba",
        "numpy",
        "setuptools",
    ],
    extras_require={
        "dev": ["pytest", "sphinx"],
    },
    entry_points={
        "console_scripts": [
            "pythermalcomfort = pythermalcomfort.cli:main",
        ],
    },
)



import math

import numpy as np
import pytest

from pythermalcomfort.classes_return import PMVPPD
from pythermalcomfort.models.pmv_ppd_iso import pmv_ppd_iso
from pythermalcomfort.utilities import Models, Units
from tests.conftest import Urls, retrieve_reference_table, validate_result


def test_pmv_ppd(get_test_url, retrieve_data) -> None:
    """Test that the function calculates the PMV and PPD model correctly for various inputs."""
    reference_table = retrieve_reference_table(
        get_test_url,
        retrieve_data,
        Urls.PMV_PPD.name,
    )
    tolerance = reference_table["tolerance"]

    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        outputs = entry["outputs"]
        # TODO change the validation table code and removed the following
        if "standard" not in inputs:
            inputs["standard"] = "iso"
        if inputs["standard"] == "iso":
            inputs["model"] = Models.iso_7730_2005.value
            del inputs["standard"]
            result = pmv_ppd_iso(**inputs)

            validate_result(result, outputs, tolerance)


class TestPmvPpd:
    """Test cases for the PMV and PPD model."""

    def test_returns_nan_for_invalid_input_values(self) -> None:
        """Test that the function returns NaN for invalid input values."""
        # Arrange
        tdb = [25, 50]
        tr = [23, 45]
        vr = [0.5, 3]
        rh = [50, 80]
        met = [1.2, 2.5]
        clo = [0.5, 1.8]

        # Act
        result = pmv_ppd_iso(
            tdb,
            tr,
            vr,
            rh,
            met,
            clo,
            model=Models.iso_7730_2005.value,
        )

        # Assert
        assert math.isnan(result.pmv[1])
        assert math.isnan(result.ppd[1])

        assert (
            round(
                pmv_ppd_iso(
                    67.28,
                    67.28,
                    0.328084,
                    86,
                    1.1,
                    1,
                    units=Units.IP.value,
                    model=Models.iso_7730_2005.value,
                ).pmv,
                1,
            )
        ) == -0.5

        np.testing.assert_equal(
            np.around(
                pmv_ppd_iso(
                    [70, 70],
                    67.28,
                    0.328084,
                    86,
                    1.1,
                    1,
                    units=Units.IP.value,
                    model=Models.iso_7730_2005.value,
                ).pmv,
                1,
            ),
            [-0.3, -0.3],
        )

        # checking that returns np.nan when outside standard applicability limits
        np.testing.assert_equal(
            pmv_ppd_iso(
                [31, 20, 20, 20, 20, 30],
                [20, 41, 20, 20, 20, 20],
                [0.1, 0.1, 2, 0.1, 0.1, 0.1],
                50,
                [1.1, 1.1, 1.1, 0.7, 1.1, 4.1],
                [0.5, 0.5, 0.5, 0.5, 2.1, 0.1],
                model=Models.iso_7730_2005.value,
            ).pmv,
            np.array([np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]),
        )

        # check results with limit_inputs disabled
        np.testing.assert_equal(
            pmv_ppd_iso(
                31,
                41,
                2,
                50,
                0.7,
                2.1,
                model=Models.iso_7730_2005.value,
                limit_inputs=False,
            ),
            PMVPPD(pmv=np.float64(2.4), ppd=np.float64(91.0), tsv="Warm"),
        )

    def test_wrong_standard(self) -> None:
        """Test that the function raises ValueError for an unsupported standard."""
        with pytest.raises(ValueError):
            pmv_ppd_iso(25, 25, 0.1, 50, 1.1, 0.5, model="random")

    def test_no_rounding(self) -> None:
        """Test that the function calculates PMV and PPD without rounding."""
        np.isclose(
            pmv_ppd_iso(
                25,
                25,
                0.1,
                50,
                1.1,
                0.5,
                round_output=False,
                model=Models.iso_7730_2005.value,
            ).pmv,
            -0.13201636,
            atol=0.01,
        )



from __future__ import annotations

import numpy as np

from pythermalcomfort.classes_input import PMVPPDInputs
from pythermalcomfort.classes_return import PMVPPD
from pythermalcomfort.models._pmv_ppd_optimized import _pmv_ppd_optimized
from pythermalcomfort.shared_functions import mapping, valid_range
from pythermalcomfort.utilities import (
    Models,
    Units,
    _check_standard_compliance_array,
    units_converter,
)


def pmv_ppd_iso(
    tdb: float | list[float],
    tr: float | list[float],
    vr: float | list[float],
    rh: float | list[float],
    met: float | list[float],
    clo: float | list[float],
    wme: float | list[float] = 0,
    model: str = Models.iso_7730_2005.value,
    units: str = Units.SI.value,
    limit_inputs: bool = True,
    round_output: bool = True,
) -> PMVPPD:
    """Return Predicted Mean Vote (PMV) and Predicted Percentage of Dissatisfied (PPD)
    calculated in accordance with the ISO 7730.

    The ISO uses the same formulation of the PMV as published by Fanger [Fanger1970]_.

    Parameters
    ----------
    tdb : float or list of floats
        Dry bulb air temperature, [°C] in [°F] if `units` = 'IP'
    tr : float or list of floats
        Mean radiant temperature, [°C] in [°F] if `units` = 'IP'
    vr : float or list of floats
        Relative air speed, [m/s] in [fps] if `units` = 'IP'

        .. note::
            vr is the relative air speed caused by body movement and not the air
            speed measured by the air speed sensor. The relative air speed is the sum of the
            average air speed measured by the sensor plus the activity-generated air speed
            (Vag). Where Vag is the activity-generated air speed caused by motion of
            individual body parts. vr can be calculated using the function
            :py:meth:`pythermalcomfort.utilities.v_relative`.

    rh : float or list of floats
        Relative humidity, [%].
    met : float or list of floats
        Metabolic rate, [met].
    clo : float or list of floats
        Clothing insulation, [clo].

        .. note::
            this is the basic insulation also known as the intrinsic clothing insulation value of the
            clothing ensemble (`I`:sub:`cl,r`), this is the thermal insulation from the skin
            surface to the outer clothing surface, including enclosed air layers, under actual
            environmental conditions. This value is not the total insulation (`I`:sub:`T,r`).
            The dynamic clothing insulation, clo, can be calculated using the function
            :py:meth:`pythermalcomfort.utilities.clo_dynamic_iso`.

    wme : float or list of floats, optional
        External work, [met]. Defaults to 0.
    model : str, optional
        Select the model you want to use to calculate the PMV. Currently, the only option available is "7730-2005".
    units : str, optional
        Select the SI (International System of Units) or the IP (Imperial Units) system.
        Supported values are 'SI' and 'IP'. Defaults to 'SI'.
    limit_inputs : bool, optional
        If True, limits the inputs to the standard applicability limits. Defaults to True.

        .. note::
            By default, if the inputs are outside the standard applicability limits the
            function returns nan. If False returns pmv and ppd values even if input values are
            outside the applicability limits of the model.

            The ISO 7730 2005 limits are 10 < tdb [°C] < 30, 10 < tr [°C] < 40,
            0 < vr [m/s] < 1, 0.8 < met [met] < 4, 0 < clo [clo] < 2, and -2 < PMV < 2.

    round_output : bool, optional
        If True, rounds output value. If False, it does not round it. Defaults to True.

    Returns
    -------
    PMVPPD
        A dataclass containing the Predicted Mean Vote and Predicted Percentage of
        Dissatisfied. See :py:class:`~pythermalcomfort.classes_return.PMVPPD` for
        more details. To access the `pmv`, `ppd`, `tsv` values, use the corresponding
        attributes of the returned `PMVPPD` instance, e.g., `result.pmv`.

    Examples
    --------
    .. code-block:: python

        from pythermalcomfort.models import pmv_ppd_iso
        from pythermalcomfort.utilities import v_relative

        tdb = 25
        tr = 25
        rh = 50
        v = 0.1
        met = 1.4
        clo_dynamic = 0.5
        # calculate relative air speed
        v_r = v_relative(v=v, met=met)
        results = pmv_ppd_iso(
            tdb=tdb,
            tr=tr,
            vr=v_r,
            rh=rh,
            met=met,
            clo=clo_dynamic,
            model="7730-2005",
        )
        print(results.pmv)  # 0.17
        print(results.ppd)  # 5.6

        result = pmv_ppd_iso(
            tdb=[22, 25],
            tr=25,
            vr=0.1,
            rh=50,
            met=1.4,
            clo=0.5,
            model="7730-2005",
        )
        print(result.pmv)  # [-0.  0.41]
        print(result.ppd)  # [5.  8.5]

    """
    # Validate inputs using the PMVPPDInputs class
    PMVPPDInputs(
        tdb=tdb,
        tr=tr,
        vr=vr,
        rh=rh,
        met=met,
        clo=clo,
        wme=wme,
        units=units,
        limit_inputs=limit_inputs,
    )

    tdb = np.array(tdb)
    tr = np.array(tr)
    rh = np.array(rh)
    vr = np.array(vr)
    met = np.array(met)
    clo = np.array(clo)
    wme = np.array(wme)

    if units.upper() == Units.IP.value:
        tdb, tr, vr = units_converter(tdb=tdb, tr=tr, v=vr)

    model = model.lower()
    if model not in [Models.iso_7730_2005.value]:
        raise ValueError(
            "PMV calculations can only be performed in compliance with ISO 7730-2005",
        )

    (
        tdb_valid,
        tr_valid,
        v_valid,
        met_valid,
        clo_valid,
    ) = _check_standard_compliance_array(
        standard=Models.iso_7730_2005.value,
        tdb=tdb,
        tr=tr,
        v=vr,
        met=met,
        clo=clo,
    )

    pmv_array = _pmv_ppd_optimized(tdb, tr, vr, rh, met, clo, wme)

    ppd_array = 100.0 - 95.0 * np.exp(
        -0.03353 * pmv_array**4.0 - 0.2179 * pmv_array**2.0,
    )

    # Checks that inputs are within the bounds accepted by the model if not return nan
    if limit_inputs:
        pmv_valid = valid_range(pmv_array, (-2, 2))  # this is the ISO limit

        all_valid = ~(
            np.isnan(tdb_valid)
            | np.isnan(tr_valid)
            | np.isnan(v_valid)
            | np.isnan(met_valid)
            | np.isnan(clo_valid)
            | np.isnan(pmv_valid)
        )
        pmv_array = np.where(all_valid, pmv_array, np.nan)
        ppd_array = np.where(all_valid, ppd_array, np.nan)

    if round_output:
        pmv_array = np.round(pmv_array, 2)
        ppd_array = np.round(ppd_array, 1)

    thermal_sensation = {
        -2.5: "Cold",
        -1.5: "Cool",
        -0.5: "Slightly Cool",
        0.5: "Neutral",
        1.5: "Slightly Warm",
        2.5: "Warm",
        10: "Hot",
    }

    return PMVPPD(
        pmv=pmv_array,
        ppd=ppd_array,
        tsv=mapping(pmv_array, thermal_sensation, right=False),
    )



from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from pythermalcomfort.classes_return import JOS3BodyParts
from pythermalcomfort.jos3_functions import construction
from pythermalcomfort.jos3_functions.construction import (
    bfb_rate,
    capacity,
    conductance,
    local_bsa,
    validate_body_parameters,
    weight_rate,
)
from pythermalcomfort.jos3_functions.matrix import (
    IDICT,
    LAYER_NAMES,
    NUM_NODES,
    index_by_layer,
    index_order,
    local_arr,
    valid_index_by_layer,
    vessel_blood_flow,
)
from pythermalcomfort.jos3_functions.parameters import Default
from pythermalcomfort.jos3_functions.thermoregulation import (
    ava_blood_flow,
    basal_met,
    clo_area_factor,
    conv_coef,
    dry_r,
    error_signals,
    evaporation,
    fixed_hc,
    fixed_hr,
    local_mbase,
    local_q_work,
    nonshivering,
    operative_temp,
    rad_coef,
    resp_heat_loss,
    shivering,
    skin_blood_flow,
    sum_bf,
    wet_r,
)
from pythermalcomfort.models import JOS3


def test_jos3_class() -> None:
    """Test the JOS3 class and its methods."""
    # Test for the initialization of JOS3 class
    # Instantiate JOS3 class
    model = JOS3()
    # Check if the object is of type JOS3
    assert isinstance(model, JOS3)

    # Test for the methods of JOS3 class
    # Instantiate the JOS3 class
    model = JOS3()

    # Call the simulate method
    model.simulate(times=60)

    # Test the reset_setpt method
    result = model._reset_setpt()
    result = result.__dict__
    assert "t_core" in result
    assert "t_skin" in result
    # Check if the attributes of the JOS3 object have been updated as expected
    assert model.to is not None
    assert np.all(model.rh == 50)
    assert np.all(model.v == 0.1)
    assert np.all(model.clo == 0)
    assert model.par == 1.25
    # Check if the new set-point temperatures for core and skin have been set
    assert np.all(model.cr_set_point == model.t_core)
    assert np.all(model.sk_set_point == model.t_skin)

    # Test: JOS3 class with sex="female"
    model = JOS3(sex="female", bmr_equation="japanese")
    model.simulate(times=60)
    # Check if mean skin temperature is not NaN
    dict_output = model.dict_results()
    assert not np.isnan(dict_output["t_skin_mean"]).any()

    # Test the simulate method with different parameters
    model = JOS3(height=1.7, weight=60, age=30)
    # Set the first phase
    model.to = 28  # Operative temperature [°C]
    model.rh = 40  # Relative humidity [%]
    model.v = 0.2  # Air velocity [m/s]
    model.par = 1.2  # Physical activity ratio [-]
    model.simulate(60)  # Exposure time = 60 [min]

    # Set the next condition
    model.to = 20  # Change only operative temperature
    model.simulate(60)  # Additional exposure time = 60 [min]

    # Define output dictionary
    dict_output = model.dict_results()

    # Results to compare (example data in csv file)
    # Get the absolute path of the current script
    current_script_path = Path(__file__).resolve()
    # Get the project directory by going up two levels from the current script path
    project_directory = Path(current_script_path).parent.parent
    # Specify the relative path to the CSV file
    relative_path = (
        Path("examples") / "jos3_output_example" / "jos3_example1 (default output).csv"
    )
    # Generate the absolute path by combining the project and the relative path
    file_path = project_directory / Path(relative_path)

    # Read the 't_skin_mean' column from the CSV file into a DataFrame
    df_t_skin_mean = (
        pd.read_csv(file_path, skiprows=2)["mean skin temperature"]
    ).tolist()

    # Check if the simulate method returns the expected type (e.g., dict)
    assert isinstance(dict_output, dict)
    assert np.allclose(dict_output["t_skin_mean"], df_t_skin_mean)

    # Call the _run method
    model._run(dtime=60000, passive=True)
    result = model.dict_results()
    # Check if the _run method returns the expected type (e.g., dict)
    assert isinstance(result, dict)

    # Test: Check property
    # Instantiate the JOS3 class
    model = JOS3()
    # Access the t_core property
    t_core = model.t_core
    # Check if t_core is of the expected type (e.g., numpy.ndarray)
    assert isinstance(t_core, np.ndarray)

    # Test with value out of range
    with pytest.raises(ValueError):
        JOS3(weight=1)

    # Test with value out of range
    with pytest.raises(ValueError):
        JOS3(age=1)

    # Test with value out of range
    with pytest.raises(ValueError):
        JOS3(height=0.1)

    # Test with value out of range
    with pytest.raises(ValueError):
        JOS3(fat=91)


def test_body_parameters() -> None:
    """Test the validate_body_parameters function."""
    # Test with valid parameters
    validate_body_parameters(height=1.75, weight=70.0, age=30, body_fat=15)
    # Test with invalid height
    with pytest.raises(ValueError):
        validate_body_parameters(height=0.1, weight=70.0, age=30, body_fat=15)
    # Test with invalid weight
    with pytest.raises(ValueError):
        validate_body_parameters(height=1.75, weight=210.0, age=30, body_fat=15)
    # Test with invalid age
    with pytest.raises(ValueError):
        validate_body_parameters(height=1.75, weight=210.0, age=101, body_fat=15)
    # Test with invalid body fat
    with pytest.raises(ValueError):
        validate_body_parameters(height=1.75, weight=210.0, age=101, body_fat=91)


def test_to17array() -> None:
    """Test the to_array_body_parts function from construction.py."""
    # Test with integer input
    result = construction.to_array_body_parts(5)
    assert isinstance(result, np.ndarray)
    assert result.shape == (17,)
    assert np.all(result == 5)

    # Test with float input
    result = construction.to_array_body_parts(5.5)
    assert isinstance(result, np.ndarray)
    assert result.shape == (17,)
    assert np.all(result == 5.5)

    # Test with list input
    result: np.ndarray = construction.to_array_body_parts(list(range(17)))
    assert isinstance(result, np.ndarray)
    assert result.shape == (17,)
    assert np.all(result == np.arange(17))

    # Test with ndarray input
    result = construction.to_array_body_parts(np.arange(17))
    assert isinstance(result, np.ndarray)
    assert result.shape == (17,)
    assert np.all(result == np.arange(17))

    # Test with dict input
    dict_input = {name: i for i, name in enumerate(JOS3BodyParts.get_attribute_names())}
    result = construction.to_array_body_parts(dict_input)
    assert isinstance(result, np.ndarray)
    assert result.shape == (17,)
    assert np.all(result == np.arange(17))

    # Test with list input of wrong length
    with pytest.raises(ValueError):
        construction.to_array_body_parts(list(range(16)))

    # Test with ndarray input of wrong length
    with pytest.raises(ValueError):
        construction.to_array_body_parts(np.arange(16))

    # Test with unsupported input type
    with pytest.raises(ValueError):
        construction.to_array_body_parts("unsupported")


def test_bsa_rate() -> None:
    """Test the bsa_rate function from construction.py."""
    # Test with default parameters
    expected_result = (
        1.0  # Since height and weight are set to default values, bsa_rate should be 1.0
    )
    result = construction.bsa_rate(height=1.72, weight=74.43, bsa_equation="dubois")
    assert result == pytest.approx(
        expected_result,
        rel=1e-3,
    )  # a relative tolerance of 1e-3

    # Test with custom parameters
    result = construction.bsa_rate(height=1.8, weight=80, bsa_equation="dubois")
    assert isinstance(result, float)

    # Test with invalid formula
    with pytest.raises(ValueError):
        construction.bsa_rate(bsa_equation="sushi", height=1.72, weight=74.43)

    # Test with non-numeric height
    with pytest.raises(TypeError):
        construction.bsa_rate(height="non-numeric", weight=74.43, bsa_equation="dubois")

    # Test with non-numeric weight
    with pytest.raises(TypeError):
        construction.bsa_rate(weight="non-numeric", height=1.72, bsa_equation="dubois")


def test_local_bsa() -> None:
    """Test the local_bsa function from construction.py."""
    # Test with default parameters
    result = local_bsa(height=1.72, weight=74.43, bsa_equation="dubois")
    assert isinstance(result, np.ndarray)
    assert result.shape == (17,)

    # Test with custom parameters
    result = local_bsa(height=1.8, weight=80, bsa_equation="dubois")
    assert isinstance(result, np.ndarray)
    assert result.shape == (17,)

    # Test with invalid formula
    with pytest.raises(ValueError):
        local_bsa(bsa_equation="unknown", height=1.72, weight=74.43)

    # Test with non-numeric height
    with pytest.raises(TypeError):
        local_bsa(height="non-numeric", weight=74.43, bsa_equation="dubois")

    # Test with non-numeric weight
    with pytest.raises(TypeError):
        local_bsa(weight="non-numeric", height=1.72, bsa_equation="dubois")


def test_weight_rate() -> None:
    """Test the weight_rate function from construction.py."""
    # Test with default parameters
    result = weight_rate(weight=74.43)
    expected_result = (
        1.0  # Since height and weight are set to default values, bsa_rate should be 1.0
    )
    assert result == pytest.approx(
        expected_result,
        rel=1e-3,
    )  # a relative tolerance of 1e-3

    # Test with custom weight
    weight = 80.0  # A valid weight value in kg
    result = weight_rate(weight=weight)
    standard_weight = 74.43
    assert (
        result == weight / standard_weight
    )  # The result should be the input weight divided by the standard weight

    # Test with non-numeric weight
    with pytest.raises(TypeError):
        weight_rate(weight="non-numeric")


def test_bfb_rate() -> None:
    """Test the bfb_rate function from construction.py."""
    # Test with default parameters
    result = bfb_rate(height=1.72, weight=74.43, bsa_equation="dubois", age=20, ci=2.59)
    assert isinstance(result, float)

    # Test with custom parameters
    result = bfb_rate(height=1.8, weight=80, bsa_equation="dubois", age=30, ci=2.7)
    assert isinstance(result, float)

    # Test with different ages
    result_young = bfb_rate(
        age=40,
        height=1.72,
        weight=74.43,
        bsa_equation="dubois",
        ci=2.59,
    )
    result_old = bfb_rate(
        age=60,
        height=1.72,
        weight=74.43,
        bsa_equation="dubois",
        ci=2.59,
    )
    assert result_old < result_young  # The BFB rate should decrease with age

    # Test with invalid equation
    with pytest.raises(ValueError):
        bfb_rate(bsa_equation="unknown", height=1.72, weight=74.43, age=20, ci=2.59)

    # Test with non-numeric height
    with pytest.raises(TypeError):
        bfb_rate(
            height="non-numeric",
            weight=74.43,
            age=20,
            ci=2.59,
            bsa_equation="dubois",
        )

    # Test with non-numeric weight
    with pytest.raises(TypeError):
        bfb_rate(
            weight="non-numeric",
            height=1.72,
            age=20,
            ci=2.59,
            bsa_equation="dubois",
        )

    # Test with non-numeric age
    with pytest.raises(TypeError):
        bfb_rate(
            age="non-numeric",
            height=1.72,
            weight=74.43,
            ci=2.59,
            bsa_equation="dubois",
        )

    # Test with non-numeric ci
    with pytest.raises(TypeError):
        bfb_rate(
            ci="non-numeric",
            height=1.72,
            weight=74.43,
            age=20,
            bsa_equation="dubois",
        )


def test_conductance() -> None:
    """Test the conductance function from construction.py."""
    # Test with default parameters
    result = conductance(height=1.72, weight=74.43, bsa_equation="dubois", fat=15.0)
    assert isinstance(result, np.ndarray)

    # Test with custom parameters
    result = conductance(height=1.8, weight=80, bsa_equation="dubois", fat=20.0)
    assert isinstance(result, np.ndarray)

    # Test with different fat rates
    result_low_fat = conductance(
        height=1.72,
        weight=74.43,
        bsa_equation="dubois",
        fat=10.0,
    )
    result_high_fat = conductance(
        height=1.72,
        weight=74.43,
        bsa_equation="dubois",
        fat=30.0,
    )
    assert np.any(
        result_low_fat != result_high_fat,
    )  # The conductance matrix should differ

    # Test with non-numeric height
    with pytest.raises(TypeError):
        conductance(height="non-numeric", weight=74.43, bsa_equation="dubois", fat=10.0)

    # Test with non-numeric weight
    with pytest.raises(TypeError):
        conductance(height=1.72, weight="non-numeric", bsa_equation="dubois", fat=10.0)

    # Test with non-numeric fat
    with pytest.raises(TypeError):
        conductance(height=1.72, weight=74.43, fat="non-numeric", bsa_equation="dubois")


def test_capacity() -> None:
    """Test the capacity function from construction.py."""
    # Test with default parameters
    result = capacity(height=1.72, weight=74.43, bsa_equation="dubois", age=20, ci=2.59)
    assert isinstance(result, np.ndarray)
    assert result.shape[0] == NUM_NODES

    # Test with non-default parameters
    result = capacity(height=1.8, weight=80, bsa_equation="takahira", age=25, ci=3)
    assert isinstance(result, np.ndarray)
    assert result.shape[0] == NUM_NODES

    # Test with invalid equation
    with pytest.raises(ValueError):
        capacity(
            bsa_equation="invalid_equation",
            height=1.72,
            weight=74.43,
            age=20,
            ci=2.59,
        )


def test_index_order() -> None:
    """Test the index_order function from matrix.py."""
    index_dict, order_count = index_order()

    # Test that output is of correct type
    assert isinstance(index_dict, dict)
    assert isinstance(order_count, int)

    # Test that output has correct keys and structure
    for key in JOS3BodyParts.get_attribute_names():
        assert key in index_dict
        for sub_key in LAYER_NAMES:
            assert sub_key in index_dict[key]

    # Test that "CB" key is present and it's value is 0
    assert "CB" in index_dict
    assert index_dict["CB"] == 0

    # Test order_count is correct
    total_layers = sum(
        1
        for bn in JOS3BodyParts.get_attribute_names()
        for ln in LAYER_NAMES
        if index_dict[bn][ln] is not None
    )
    assert total_layers + 1 == order_count  # +1 because of the "CB" key


def test_index_by_layer() -> None:
    """Test the index_by_layer function from matrix.py."""
    # Test that output is of correct type and length for each layer
    for layer in LAYER_NAMES:
        indices = index_by_layer(layer)
        assert isinstance(indices, list)
        assert len(indices) == sum(
            1
            for bn in JOS3BodyParts.get_attribute_names()
            if IDICT[bn][layer] is not None
        )

    # Test that each index is in the correct range
    for layer in LAYER_NAMES:
        indices = index_by_layer(layer)
        for index in indices:
            assert (
                0
                <= index
                < sum(
                    1
                    for bn in JOS3BodyParts.get_attribute_names()
                    for ln in LAYER_NAMES
                    if IDICT[bn][ln] is not None
                )
                + 1
            )  # +1 because of the "CB" key


def test_valid_index_by_layer() -> None:
    """Test the valid_index_by_layer function from matrix.py."""
    # Test that output is of correct type and length for each layer
    for layer in LAYER_NAMES:
        indices = valid_index_by_layer(layer)
        assert isinstance(indices, list)
        assert len(indices) == sum(
            1
            for bn in JOS3BodyParts.get_attribute_names()
            if IDICT[bn][layer] is not None
        )

    # Test that each index is in the correct range
    for layer in LAYER_NAMES:
        indices = valid_index_by_layer(layer)
        for index in indices:
            assert 0 <= index < len(JOS3BodyParts.get_attribute_names())

    # Test for incorrect layer input
    with pytest.raises(KeyError):
        valid_index_by_layer("non_existent_layer")


def test_local_arr() -> None:
    """Test the local_arr function from matrix.py."""
    # Initialize some random blood flow values
    rng = np.random.default_rng(1337)
    bf_core = rng.random(len(JOS3BodyParts.get_attribute_names()))
    bf_muscle = rng.random(len(JOS3BodyParts.get_attribute_names()))
    bf_fat = rng.random(len(JOS3BodyParts.get_attribute_names()))
    bf_skin = rng.random(len(JOS3BodyParts.get_attribute_names()))
    bf_ava_hand = rng.random(1)[0]
    bf_ava_foot = rng.random(1)[0]

    # Call the function with these values
    result = local_arr(bf_core, bf_muscle, bf_fat, bf_skin, bf_ava_hand, bf_ava_foot)

    # Check that the result is a 2D array with the correct dimensions
    assert isinstance(result, np.ndarray)
    assert result.shape == (NUM_NODES, NUM_NODES)

    # Check that the elements in the array are correct
    # You could add more specific checks here, depending on the expected
    # behavior of the function
    for i, bn in enumerate(JOS3BodyParts.get_attribute_names()):
        index_of = IDICT[bn]

        assert np.isclose(
            result[index_of["core"], index_of["artery"]],
            1.067 * bf_core[i],
        )
        assert np.isclose(
            result[index_of["skin"], index_of["artery"]],
            1.067 * bf_skin[i],
        )
        assert np.isclose(
            result[index_of["vein"], index_of["core"]],
            1.067 * bf_core[i],
        )
        assert np.isclose(
            result[index_of["vein"], index_of["skin"]],
            1.067 * bf_skin[i],
        )

        if index_of["muscle"] is not None:
            assert np.isclose(
                result[index_of["muscle"], index_of["artery"]],
                1.067 * bf_muscle[i],
            )
            assert np.isclose(
                result[index_of["vein"], index_of["muscle"]],
                1.067 * bf_muscle[i],
            )
        if index_of["fat"] is not None:
            assert np.isclose(
                result[index_of["fat"], index_of["artery"]],
                1.067 * bf_fat[i],
            )
            assert np.isclose(
                result[index_of["vein"], index_of["fat"]],
                1.067 * bf_fat[i],
            )

        if i in [7, 10]:
            assert np.isclose(
                result[index_of["sfvein"], index_of["artery"]],
                1.067 * bf_ava_hand,
            )
        if i in [13, 16]:
            assert np.isclose(
                result[index_of["sfvein"], index_of["artery"]],
                1.067 * bf_ava_foot,
            )


def test_vessel_blood_flow() -> None:
    """Test the vessel_blood_flow function from matrix.py."""
    # Initialize some random blood flow values
    rng = np.random.default_rng(1337)
    bf_core = rng.random(17)
    bf_muscle = rng.random(17)
    bf_fat = rng.random(17)
    bf_skin = rng.random(17)
    bf_ava_hand = rng.random(1)[0]
    bf_ava_foot = rng.random(1)[0]

    # Call the function with these values
    bf_art, bf_vein = vessel_blood_flow(
        bf_core,
        bf_muscle,
        bf_fat,
        bf_skin,
        bf_ava_hand,
        bf_ava_foot,
    )

    # Check that the result is a 1D array with the correct dimensions
    assert isinstance(bf_art, np.ndarray)
    assert isinstance(bf_vein, np.ndarray)
    assert bf_art.shape == (17,)
    assert bf_vein.shape == (17,)

    # Calculate expected blood flow
    xbf = bf_core + bf_muscle + bf_fat + bf_skin

    # Check that the elements in the array are correct
    assert np.isclose(bf_art[0], xbf[0])
    assert np.isclose(bf_vein[0], xbf[0])

    assert np.isclose(bf_art[1], xbf[1] + xbf[0])
    assert np.isclose(bf_vein[1], xbf[1] + xbf[0])

    assert np.isclose(bf_art[2], xbf[2])
    assert np.isclose(bf_vein[2], xbf[2])

    assert np.isclose(bf_art[3], xbf[3])
    assert np.isclose(bf_vein[3], xbf[3])


# test for thermoregulation.py
def test_conv_coef() -> None:
    """Test the conv_coef function for calculating convective coefficients based on posture."""
    # Test case 1: Default values
    hc_expected = np.array(
        [
            4.48,
            4.48,
            2.97,
            2.91,
            2.85,
            3.61,
            3.55,
            3.67,
            3.61,
            3.55,
            3.67,
            2.80,
            2.04,
            2.04,
            2.80,
            2.04,
            2.04,
        ],
    )
    assert np.array_equal(
        conv_coef(posture="standing", v=0.1, tdb=28.8, t_skin=34),
        hc_expected,
    )

    # Test case 2: Sitting posture
    hc_expected = np.array(
        [
            4.75,
            4.75,
            3.12,
            2.48,
            1.84,
            3.76,
            3.62,
            2.06,
            3.76,
            3.62,
            2.06,
            2.98,
            2.98,
            2.62,
            2.98,
            2.98,
            2.62,
        ],
    )
    assert np.array_equal(
        conv_coef(posture="sitting", tdb=28.8, v=0.1, t_skin=34.0),
        hc_expected,
    )

    # Test case 3: Invalid posture
    with pytest.raises(ValueError):
        conv_coef(posture="invalid", tdb=28.8, v=0.1, t_skin=34.0)

    # Test case 4: Lying posture with different tdb and t_skin values
    tdb = np.full(17, 28.8)
    t_skin = np.full(17, 34.0)
    hc_expected = np.array(
        [
            1.105,
            1.105,
            1.211,
            1.211,
            1.211,
            0.913,
            2.081,
            2.178,
            0.913,
            2.081,
            2.178,
            0.945,
            0.385,
            0.200,
            0.945,
            0.385,
            0.200,
        ],
    ) * (
        abs(tdb - t_skin)
        ** np.array(
            [
                0.345,
                0.345,
                0.046,
                0.046,
                0.046,
                0.373,
                0.850,
                0.297,
                0.373,
                0.850,
                0.297,
                0.447,
                0.580,
                0.966,
                0.447,
                0.580,
                0.966,
            ],
        )
    )
    assert np.allclose(
        conv_coef(posture="lying", tdb=tdb, t_skin=t_skin, v=0.1),
        hc_expected,
    )

    # Test case 5: Forced convection (v > 0.2)
    v = np.full(17, 0.3)
    hc_expected = np.array(
        [
            15.0,
            15.0,
            11.0,
            17.0,
            13.0,
            17.0,
            17.0,
            20.0,
            17.0,
            17.0,
            20.0,
            14.0,
            15.8,
            15.1,
            14.0,
            15.8,
            15.1,
        ],
    ) * (
        v
        ** np.array(
            [
                0.62,
                0.62,
                0.67,
                0.49,
                0.60,
                0.59,
                0.61,
                0.60,
                0.59,
                0.61,
                0.60,
                0.61,
                0.74,
                0.62,
                0.61,
                0.74,
                0.62,
            ],
        )
    )
    assert np.allclose(
        conv_coef(v=v, tdb=28.8, posture="standing", t_skin=34.0),
        hc_expected,
    )


def test_rad_coef() -> None:
    """Test the rad_coef function for calculating radiative coefficients based on posture."""
    # Test with valid postures
    valid_postures = {
        "standing": np.array(
            [
                4.89,
                4.89,
                4.32,
                4.09,
                4.32,
                4.55,
                4.43,
                4.21,
                4.55,
                4.43,
                4.21,
                4.77,
                5.34,
                6.14,
                4.77,
                5.34,
                6.14,
            ],
        ),
        "sitting": np.array(
            [
                4.96,
                4.96,
                3.99,
                4.64,
                4.21,
                4.96,
                4.21,
                4.74,
                4.96,
                4.21,
                4.74,
                4.10,
                4.74,
                6.36,
                4.10,
                4.74,
                6.36,
            ],
        ),
        "lying": np.array(
            [
                5.475,
                5.475,
                3.463,
                3.463,
                3.463,
                4.249,
                4.835,
                4.119,
                4.249,
                4.835,
                4.119,
                4.440,
                5.547,
                6.085,
                4.440,
                5.547,
                6.085,
            ],
        ),
        "sedentary": np.array(
            [
                4.96,
                4.96,
                3.99,
                4.64,
                4.21,
                4.96,
                4.21,
                4.74,
                4.96,
                4.21,
                4.74,
                4.10,
                4.74,
                6.36,
                4.10,
                4.74,
                6.36,
            ],
        ),
        "supine": np.array(
            [
                5.475,
                5.475,
                3.463,
                3.463,
                3.463,
                4.249,
                4.835,
                4.119,
                4.249,
                4.835,
                4.119,
                4.440,
                5.547,
                6.085,
                4.440,
                5.547,
                6.085,
            ],
        ),
    }
    for posture, expected in valid_postures.items():
        assert np.allclose(rad_coef(posture=posture), expected)

    # Test with invalid postures
    invalid_postures = ["invalid", "running", "jumping"]
    for posture in invalid_postures:
        with pytest.raises(ValueError):
            rad_coef(posture=posture)


def test_fixed_hc() -> None:
    """Test the fixed_hc function for calculating fixed convective heat transfer coefficient."""
    hc = np.ones(17) * 3
    v = np.ones(17) * 0.1

    # Call the fixed_hc function
    fixed_hc_values = fixed_hc(hc=hc, v=v)

    # Check if the function returns a numpy array
    assert isinstance(fixed_hc_values, np.ndarray)

    # Check if the function returns an array of the same shape as input
    assert fixed_hc_values.shape == hc.shape

    # Check if the function returns expected values (based on the formula)
    mean_hc = np.average(hc, weights=Default.local_bsa)
    mean_va = np.average(v, weights=Default.local_bsa)
    mean_hc_whole = max(3, 8.600001 * (mean_va**0.53))
    expected_fixed_hc = hc * mean_hc_whole / mean_hc
    assert np.allclose(fixed_hc_values, expected_fixed_hc)


def test_fixed_hr() -> None:
    """Test the fixed_hr function for calculating radiative coefficient."""
    hr = np.ones(17) * 5

    # Call the fixed_hr function
    fixed_hr_values = fixed_hr(hr)

    # Check if the function returns a numpy array
    assert isinstance(fixed_hr_values, np.ndarray)

    # Check if the function returns an array of the same shape as input
    assert fixed_hr_values.shape == hr.shape

    # Check if the function returns expected values (based on the formula)
    mean_hr = np.average(hr, weights=Default.local_bsa)
    expected_fixed_hr = hr * 4.7 / mean_hr
    assert np.allclose(fixed_hr_values, expected_fixed_hr)


def test_operative_temp() -> None:
    """Test the operative_temp function for calculating operative temperature."""
    # Test with scalar inputs
    tdb = 25.0
    tr = 25.0
    hc = 3
    hr = 5
    to = operative_temp(tdb, tr, hc, hr)
    expected_to = 25.0
    assert (
        pytest.approx(to) == expected_to
    )  # Since tdb and tr are the same, to should also be the same

    # Test with array inputs
    tdb = np.array([20.0, 22.0, 24.0])
    tr = np.array([25.0, 27.0, 29.0])
    hc = np.array([0.5, 0.6, 0.7])
    hr = np.array([0.5, 0.4, 0.3])
    to = operative_temp(tdb, tr, hc, hr)
    expected_to = (hc * tdb + hr * tr) / (hc + hr)
    assert np.allclose(to, expected_to)

    # Test with mixed scalar and array inputs
    tdb = 20.0
    tr = np.array([25.0, 27.0, 29.0])
    hc = 0.5
    hr = np.array([0.5, 0.4, 0.3])
    to = operative_temp(tdb, tr, hc, hr)
    expected_to = (hc * tdb + hr * tr) / (hc + hr)
    assert np.allclose(to, expected_to)


def test_clo_area_factor() -> None:
    """Test the clo_area_factor function for calculating clothing area factor."""
    # Test with single value less than 0.5
    clo = 0.4
    expected_result = 1.08
    assert clo_area_factor(clo) == pytest.approx(expected_result, rel=1e-3)

    # Test with single value less than 0.5
    clo = 0.6
    expected_result = 1.11
    assert clo_area_factor(clo) == pytest.approx(expected_result, rel=1e-3)

    # Test with zero value
    clo = 0
    expected_result = 1  # clothing area factor should be 1 when clo is 0
    assert clo_area_factor(clo) == pytest.approx(expected_result, rel=1e-3)

    # Test with array values
    clo = np.array([0.4, 0.6, 0.2])
    expected_result = np.array([1.08, 1.11, 1.04])
    np.testing.assert_allclose(clo_area_factor(clo), expected_result, rtol=1e-3)


def test_dry_r() -> None:
    """Test the dry_r function for calculating dry resistance."""
    # Test with single values
    hc, hr, clo = 3, 5, 0.5
    expected_result = 0.191
    assert dry_r(hc, hr, clo) == pytest.approx(expected_result, rel=1e-3)

    # Test with array values
    hc = np.array([3, 3])
    hr = np.array([5, 5])
    clo = np.array([0.5, 0.5])
    expected_result = np.array([0.191, 0.191])
    assert dry_r(hc, hr, clo) == pytest.approx(expected_result, rel=1e-3)

    # Test with zero values
    hc, hr, clo = 0, 0, 0
    with pytest.raises(ZeroDivisionError):
        dry_r(hc, hr, clo)

    # Test negative hc and hr values
    hc, hr, clo = -10.0, -5.0, 1.0
    with pytest.raises(ValueError):
        dry_r(hc, hr, clo)


def test_wet_r() -> None:
    """Test the wet_r function for calculating wet resistance."""
    # Test with single values
    hc = 10.0
    clo = 0.5
    i_clo = 0.45
    lewis_rate = 16.5
    expected_result = 0.01594
    assert wet_r(hc, clo, i_clo, lewis_rate) == pytest.approx(expected_result, rel=1e-3)

    # Test with array values
    hc = np.array([10.0, 10.0])
    clo = np.array([0.5, 0.5])
    i_clo = np.array([0.45, 0.45])
    lewis_rate = 16.5
    expected_result = np.array(
        [0.01594, 0.01594],
    )  # Replace with actual expected result based on your function's logic
    np.testing.assert_allclose(
        wet_r(hc, clo, i_clo, lewis_rate),
        expected_result,
        rtol=1e-3,
    )

    # Test with zero values
    hc, clo = 0, 0
    with pytest.raises(ZeroDivisionError):
        wet_r(hc, clo, i_clo, lewis_rate)

    # Test with zero value for clo
    hc, clo = -1, 0.5
    with pytest.raises(ValueError):
        wet_r(hc, clo, i_clo, lewis_rate)


def test_error_signals() -> None:
    """Test the error_signals function for calculating error signals."""
    # Test with default value:
    wrms, clds = error_signals()
    assert wrms == 0
    assert clds == 0

    # Test with positive value
    wrms, clds = error_signals(2)
    assert wrms > 0
    assert clds == 0

    # Test with negative value
    wrms, clds = error_signals(-2)
    assert wrms == 0
    assert clds > 0

    # Test with array
    err_sk = np.array(
        [-2, 2, -3, 3, -1, 1, 0, -0.5, 0.5, -2, 2, -1, 1, -0.5, 0.5, -1, 1],
    )
    wrms, clds = error_signals(err_sk)
    assert wrms > 0
    assert clds > 0

    # Test with wrong length array
    err_sk = np.array([-2, 2, -3])
    with pytest.raises(ValueError):
        error_signals(err_sk)


def test_evaporation() -> None:
    """Test the evaporation function for calculating evaporation rate."""
    # Test with basic parameters
    err_cr = np.array([0.5])
    err_sk = np.array(
        [
            0.2,
            0.3,
            0.4,
            0.5,
            0.6,
            0.7,
            0.8,
            0.9,
            1.0,
            1.1,
            1.2,
            1.3,
            1.4,
            1.5,
            1.6,
            1.7,
            1.8,
        ],
    )
    t_skin = np.array([34.0] * 17)
    tdb = np.array([25.0] * 17)
    rh = np.array([50.0] * 17)
    ret = np.array([0.01] * 17)

    wet, e_sk, e_max, e_sweat = evaporation(
        err_cr,
        err_sk,
        t_skin,
        tdb,
        rh,
        ret,
        height=1.72,
        weight=74.43,
        bsa_equation="dubois",
        age=20,
    )

    assert isinstance(wet, np.ndarray)
    assert isinstance(e_sk, np.ndarray)
    assert isinstance(e_max, np.ndarray)
    assert isinstance(e_sweat, np.ndarray)
    assert np.all(wet >= 0)
    assert np.all(wet <= 1)

    # Test age effects
    err_cr = np.array([0.5])
    err_sk = np.array([0.2] * 17)
    t_skin = np.array([34.0] * 17)
    tdb = np.array([25.0] * 17)
    rh = np.array([50.0] * 17)
    ret = np.array([0.01] * 17)

    wet_young, e_sk_young, e_max_young, e_sweat_young = evaporation(
        err_cr,
        err_sk,
        t_skin,
        tdb,
        rh,
        ret,
        age=20,
        height=1.72,
        weight=74.43,
        bsa_equation="dubois",
    )
    wet_old, e_sk_old, e_max_old, e_sweat_old = evaporation(
        err_cr,
        err_sk,
        t_skin,
        tdb,
        rh,
        ret,
        age=65,
        height=1.72,
        weight=74.43,
        bsa_equation="dubois",
    )

    assert np.all(e_sweat_old <= e_sweat_young)

    # Test with valid BSA equations
    err_cr = np.array([0.5])
    err_sk = np.array([0.2] * 17)
    t_skin = np.array([34.0] * 17)
    tdb = np.array([25.0] * 17)
    rh = np.array([50.0] * 17)
    ret = np.array([0.01] * 17)

    valid_bsa_equations_list = ["dubois", "takahira", "fujimoto", "kurazumi"]
    for bsa_equation in valid_bsa_equations_list:
        wet, e_sk, e_max, e_sweat = evaporation(
            err_cr,
            err_sk,
            t_skin,
            tdb,
            rh,
            ret,
            bsa_equation=bsa_equation,
            height=1.72,
            weight=74.43,
            age=20,
        )

        assert isinstance(wet, np.ndarray)
        assert isinstance(e_sk, np.ndarray)
        assert isinstance(e_max, np.ndarray)
        assert isinstance(e_sweat, np.ndarray)

    # Test with invalid BSA equation
    invalid_bsa_equation = "sushi"
    with pytest.raises(ValueError):
        evaporation(
            err_cr,
            err_sk,
            t_skin,
            tdb,
            rh,
            ret,
            bsa_equation=invalid_bsa_equation,
            height=1.72,
            weight=74.43,
            age=20,
        )

    # Test to ensure no errors occur when e_max is zero
    t_skin = np.array([34.0] * 17)
    tdb = np.array([34.0] * 17)
    rh = np.array([100] * 17)

    # Call the evaporation function
    wet, e_sk, e_max, e_sweat = evaporation(
        err_cr,
        err_sk,
        t_skin,
        tdb,
        rh,
        ret,
        height=1.72,
        weight=74.43,
        bsa_equation="dubois",
        age=20,
    )

    expected_e_max = 0.001
    expected_wet = 1
    # Check that all elements in e_max have been replaced with 0.001
    assert np.all(
        e_max == expected_e_max,
    )  # Verify that e_max has been replaced by 0.001
    assert np.all(
        wet == pytest.approx(expected_wet, rel=1e-3),
    )  # Verify that wet is nealy 1


def test_skin_blood_flow() -> None:
    """Test the skin_blood_flow function for calculating blood flow in the skin."""
    # Test with basic values
    err_cr = np.array([0.5])
    err_sk = np.array(
        [
            0.2,
            0.3,
            0.4,
            0.5,
            0.6,
            0.7,
            0.8,
            0.9,
            1.0,
            1.1,
            1.2,
            1.3,
            1.4,
            1.5,
            1.6,
            1.7,
            1.8,
        ],
    )

    bf_skin = skin_blood_flow(
        err_cr,
        err_sk,
        height=1.72,
        weight=74.43,
        bsa_equation="dubois",
        age=20,
        ci=2.59,
    )

    assert isinstance(bf_skin, np.ndarray)
    assert np.all(bf_skin >= 0)

    # Test age effect
    err_cr = np.array([0.5])
    err_sk = np.array([0.2] * 17)

    bf_skin_young = skin_blood_flow(
        err_cr,
        err_sk,
        age=20,
        height=1.72,
        weight=74.43,
        bsa_equation="dubois",
        ci=2.59,
    )
    bf_skin_old = skin_blood_flow(
        err_cr,
        err_sk,
        age=65,
        height=1.72,
        weight=74.43,
        bsa_equation="dubois",
        ci=2.59,
    )

    assert np.all(bf_skin_old <= bf_skin_young)

    # Test valid BSA equation
    valid_bsa_equations_list = ["dubois", "takahira", "fujimoto", "kurazumi"]
    for bsa_equation in valid_bsa_equations_list:
        err_cr = np.array([0.5])
        err_sk = np.array([0.2] * 17)

        bf_skin: np.ndarray = skin_blood_flow(
            err_cr,
            err_sk,
            bsa_equation=bsa_equation,
            height=1.72,
            weight=74.43,
            age=20,
            ci=2.59,
        )

        assert isinstance(bf_skin, np.ndarray)
        assert np.all(bf_skin >= 0)

    # Test with invalid BSA equation
    invalid_bsa_equation = "sushi"
    with pytest.raises(ValueError):
        skin_blood_flow(
            err_cr,
            err_sk,
            bsa_equation=invalid_bsa_equation,
            height=1.72,
            weight=74.43,
            age=20,
            ci=2.59,
        )


def test_ava_blood_flow() -> None:
    """Test the ava_blood_flow function for calculating blood flow in arteriovenous anastomoses (AVA)."""
    # Test with basic parameters
    err_cr = np.array([0.5, 0.6, 0.7, 0.8, 0.9])
    err_sk = np.array(
        [
            0.2,
            0.3,
            0.4,
            0.5,
            0.6,
            0.7,
            0.8,
            0.9,
            1.0,
            1.1,
            1.2,
            1.3,
            1.4,
            1.5,
            1.6,
            1.7,
            1.8,
        ],
    )

    bf_ava_hand, bf_ava_foot = ava_blood_flow(
        err_cr,
        err_sk,
        height=1.72,
        weight=74.43,
        bsa_equation="dubois",
        age=20,
        ci=2.59,
    )

    assert isinstance(bf_ava_hand, float)
    assert isinstance(bf_ava_foot, float)
    assert bf_ava_hand >= 0
    assert bf_ava_foot >= 0

    # Test age effect
    err_cr = np.array([0.5, 0.6, 0.7, 0.8, 0.9])
    err_sk = np.array([0.2] * 17)

    bf_ava_hand_young, bf_ava_foot_young = ava_blood_flow(
        err_cr,
        err_sk,
        age=20,
        height=1.72,
        weight=74.43,
        bsa_equation="dubois",
        ci=2.59,
    )
    bf_ava_hand_old, bf_ava_foot_old = ava_blood_flow(
        err_cr,
        err_sk,
        age=65,
        height=1.72,
        weight=74.43,
        bsa_equation="dubois",
        ci=2.59,
    )

    assert bf_ava_hand_old <= bf_ava_hand_young
    assert bf_ava_foot_old <= bf_ava_foot_young

    # Test with input length mismatch
    err_cr = np.array([0.5, 0.6, 0.7])
    err_sk = np.array([0.2, 0.3])
    with pytest.raises(TypeError):
        ava_blood_flow(
            err_cr,
            err_sk,
            height=1.72,
            weight=74.43,
            bsa_equation="dubois",
            age=20,
            ci=2.59,
        )

    # Test with zero signal
    err_cr = np.array([0.0, 0.0, 0.0, 0.0, 0.0])
    err_sk = np.array([0.0] * 17)
    bf_ava_hand, bf_ava_foot = ava_blood_flow(
        err_cr,
        err_sk,
        height=1.72,
        weight=74.43,
        bsa_equation="dubois",
        age=20,
        ci=2.59,
    )

    # (Need to be investigated)
    # Expected results are 1.71258 for hands and 1.42223 for feet when focusing on the
    # equation (Takemori, 1995.), but they might be 0 considering
    # characteristics of ava blood flow.
    expected_result_hand = 1.71258
    expected_result_foot = 1.42223

    assert bf_ava_hand == pytest.approx(expected_result_hand, rel=1e-3)
    assert bf_ava_foot == pytest.approx(expected_result_foot, rel=1e-3)


def test_basal_met() -> None:
    """Test the basal_met function for calculating basal metabolic rate (BMR)."""
    # Test with default values
    bmr = basal_met(
        height=1.72,
        weight=74.43,
        age=20,
        sex="male",
        bmr_equation="harris-benedict",
    )
    expected_result = 87.95
    assert bmr == pytest.approx(expected_result, rel=1e-3)

    # Test with custom values
    bmr = basal_met(
        height=1.80,
        weight=70,
        age=25,
        sex="male",
        bmr_equation="harris-benedict",
    )
    expected_result = 85.66
    assert bmr == pytest.approx(expected_result, rel=1e-3)

    # Test with valid BMR equations
    valid_equations_list = [
        "harris-benedict",
        "harris-benedict_origin",
        "japanese",
        "ganpule",
    ]
    for valid_equation in valid_equations_list:
        bmr = basal_met(
            bmr_equation=valid_equation,
            height=1.72,
            weight=74.4,
            age=20,
            sex="male",
        )
        assert isinstance(bmr, float)

    # Test with invalid BMR equation
    invalid_bsa_equation = "sushi"
    with pytest.raises(ValueError):
        basal_met(
            bmr_equation=invalid_bsa_equation,
            height=1.72,
            weight=74.43,
            age=20,
            sex="male",
        )


def test_local_mbase() -> None:
    """Test the local_mbase function for calculating local metabolic rates."""
    # Test with default values
    mbase_cr, mbase_ms, mbase_fat, mbase_sk = local_mbase()
    assert isinstance(mbase_cr, np.ndarray)
    assert isinstance(mbase_ms, np.ndarray)
    assert isinstance(mbase_fat, np.ndarray)
    assert isinstance(mbase_sk, np.ndarray)

    # Test with custom values
    mbase_cr, mbase_ms, mbase_fat, mbase_sk = local_mbase(
        height=1.80,
        weight=70,
        age=25,
        sex="male",
        bmr_equation="harris-benedict",
    )

    # Check that each element of mbase_cr is greater than the corresponding elements
    # of mbase_ms, mbase_fat, and mbase_sk
    assert all(mbase_cr > mbase_ms)
    assert all(mbase_cr > mbase_fat)
    assert all(mbase_cr > mbase_sk)


def test_local_q_work() -> None:
    """Test the local_q_work function for calculating local work rate."""
    # Test with par = 1.5
    q_work = local_q_work(bmr=100, par=1.5)
    assert all(q_work >= 0)
    # Test with par = 1.0
    q_work = local_q_work(bmr=100, par=1.0)
    assert all(q_work == 0)
    # Test with par < 1.0
    with pytest.raises(ValueError):
        local_q_work(bmr=100, par=0.8)


def test_shivering() -> None:
    """Test the shivering function for calculating shivering thermogenesis."""
    # Test with zero error signals
    err_cr = np.zeros(17)
    err_sk = np.zeros(17)
    t_core = np.ones(17) * 36
    t_skin = np.ones(17) * 32
    q_shiv = shivering(
        err_cr,
        err_sk,
        t_core,
        t_skin,
        height=1.72,
        weight=74.43,
        age=20,
        sex="male",
        dtime=1,
        options=None,
        bsa_equation="dubois",
    )
    assert all(q_shiv == 0)

    # Test with positive error signals
    err_cr = np.ones(17) * 1
    err_sk = np.ones(17) * 1
    t_core = np.ones(17) * 36
    t_skin = np.ones(17) * 32
    q_shiv = shivering(
        err_cr,
        err_sk,
        t_core,
        t_skin,
        height=1.72,
        weight=74.43,
        age=20,
        sex="male",
        dtime=1,
        options=None,
        bsa_equation="dubois",
    )
    assert all(q_shiv == 0)

    # Test positive shivering thermogenesis with negative error signals
    err_cr = np.ones(17) * -1
    err_sk = np.ones(17) * -2
    t_core = np.ones(17) * 36
    t_skin = np.ones(17) * 32
    q_shiv = shivering(
        err_cr,
        err_sk,
        t_core,
        t_skin,
        height=1.72,
        weight=74.43,
        age=20,
        sex="male",
        dtime=1,
        options=None,
        bsa_equation="dubois",
    )
    assert all(q_shiv > 0)

    # Test with different age group
    err_cr = np.ones(17) * -1
    err_sk = np.ones(17) * -2
    t_core = np.ones(17) * 36
    t_skin = np.ones(17) * 32

    age_list = [25, 35, 45, 55, 65, 75, 85]
    q_shiv_by_age = {}

    for age in age_list:
        q_shiv = shivering(
            err_cr,
            err_sk,
            t_core,
            t_skin,
            age=age,
            height=1.72,
            weight=74.43,
            sex="male",
            dtime=1,
            options=None,
            bsa_equation="dubois",
        )
        q_shiv_by_age[age] = q_shiv

    # Compare the results for different ages
    for i in range(len(age_list) - 1):
        age_younger = age_list[i]
        age_older = age_list[i + 1]
        assert all(q_shiv_by_age[age_younger] > q_shiv_by_age[age_older])


def test_nonshivering() -> None:
    """Test the nonshivering function for calculating non-shivering thermogenesis."""
    # Test with zero error signals
    err_sk = np.zeros(17)
    q_nst = nonshivering(
        err_sk,
        height=1.72,
        weight=74.43,
        age=20,
        bsa_equation="dubois",
        cold_acclimation=False,
        batpositive=True,
    )
    assert all(q_nst == 0)

    # Test with zero error signals
    err_sk = np.ones(17) * 1
    q_nst = nonshivering(
        err_sk,
        height=1.72,
        weight=74.43,
        age=20,
        bsa_equation="dubois",
        cold_acclimation=False,
        batpositive=True,
    )
    assert all(q_nst == 0)

    # Test with negative error signals
    err_sk = np.ones(17) * -1
    q_nst = nonshivering(
        err_sk,
        height=1.72,
        weight=74.43,
        age=20,
        bsa_equation="dubois",
        cold_acclimation=False,
        batpositive=True,
    )
    assert all(q_nst >= 0)

    # Test age effect on BAT (brown adipose tissue) that affects NST limit
    err_sk = np.ones(17) * -10  # Set -10 to check the NST limit is working
    age_list = [25, 35, 45]
    q_nst_by_age = {}
    sum_q_nst_by_age = {}

    for age in age_list:
        q_nst = nonshivering(
            err_sk,
            age=age,
            height=1.72,
            weight=74.43,
            bsa_equation="dubois",
        )
        q_nst_by_age[age] = q_nst
        sum_q_nst_by_age[age] = np.sum(q_nst_by_age[age])

    # Compare the results for different ages
    for i in range(len(age_list) - 1):
        age_younger = age_list[i]
        age_older = age_list[i + 1]
        assert sum_q_nst_by_age[age_younger] > sum_q_nst_by_age[age_older]


def test_cold_acclimation_non_shivering() -> None:
    """Test the cold acclimation effect on non-shivering thermogenesis."""
    # Test cold acclimation that affects NST limit
    err_sk = np.ones(17) * -10  # Set -10 to check the NST limit is working
    q_nst_no_acclimation = nonshivering(
        err_sk,
        cold_acclimation=False,
        batpositive=True,
        height=1.72,
        weight=74.43,
        age=20,
        bsa_equation="dubois",
    )
    q_nst_with_acclimation = nonshivering(
        err_sk,
        cold_acclimation=True,
        batpositive=True,
        height=1.72,
        weight=74.43,
        age=20,
        bsa_equation="dubois",
    )
    assert not np.array_equal(q_nst_no_acclimation, q_nst_with_acclimation)


def test_sum_bf() -> None:
    """Test the sum_bf function for calculating body fat."""
    # Test to check output type
    bf_core = np.array([1, 2, 3, 4])
    bf_muscle = np.array([1, 2, 3, 4])
    bf_fat = np.array([1, 2, 3, 4])
    bf_skin = np.array([1, 2, 3, 4])
    bf_ava_hand = 1.0
    bf_ava_foot = 1.0

    co = sum_bf(bf_core, bf_muscle, bf_fat, bf_skin, bf_ava_hand, bf_ava_foot)
    assert isinstance(co, float)

    # Test with custom values 1
    bf_core = np.array([1, 2, 3, 4])
    bf_muscle = np.array([1, 2, 3, 4])
    bf_fat = np.array([1, 2, 3, 4])
    bf_skin = np.array([1, 2, 3, 4])
    bf_ava_hand = 1.0
    bf_ava_foot = 1.0

    co = sum_bf(bf_core, bf_muscle, bf_fat, bf_skin, bf_ava_hand, bf_ava_foot)
    expected_co = (
        bf_core.sum()
        + bf_muscle.sum()
        + bf_fat.sum()
        + bf_skin.sum()
        + 2 * bf_ava_hand
        + 2 * bf_ava_foot
    )
    assert co == expected_co

    # Test with custom values 2
    bf_core = np.array([1, 2, 3, 4])
    bf_muscle = np.array([1, 2, 3, 4])
    bf_fat = np.array([1, 2, 3, 4])
    bf_skin = np.array([1, 2, 3, 4])
    bf_ava_hand = 1.0
    bf_ava_foot = 1.0

    bf_core_copy = bf_core.copy()
    bf_muscle_copy = bf_muscle.copy()
    bf_fat_copy = bf_fat.copy()
    bf_skin_copy = bf_skin.copy()

    sum_bf(bf_core, bf_muscle, bf_fat, bf_skin, bf_ava_hand, bf_ava_foot)

    assert np.array_equal(bf_core, bf_core_copy)
    assert np.array_equal(bf_muscle, bf_muscle_copy)
    assert np.array_equal(bf_fat, bf_fat_copy)
    assert np.array_equal(bf_skin, bf_skin_copy)


def test_resp_heat_loss() -> None:
    """Test the resp_heat_loss function for calculating respiratory heat loss."""
    # Test to check output type
    tdb = 25.0
    p_a = 1.0
    q_thermogenesis_total = 100.0

    res_sh, res_lh = resp_heat_loss(tdb, p_a, q_thermogenesis_total)
    assert isinstance(res_sh, float)
    assert isinstance(res_lh, float)

    # Test to check correct heat loss
    tdb = 25.0
    p_a = 1.0
    q_thermogenesis_total = 100.0

    res_sh, res_lh = resp_heat_loss(tdb, p_a, q_thermogenesis_total)
    expected_res_sh = 0.0014 * q_thermogenesis_total * (34 - tdb)
    expected_res_lh = 0.0173 * q_thermogenesis_total * (5.87 - p_a)

    assert res_sh == expected_res_sh
    assert res_lh == expected_res_lh

    # Test sensible heat loss at 34oC of air temp
    tdb = 34.0
    p_a = 1.0
    q_thermogenesis_total = 100.0

    res_sh, _ = resp_heat_loss(tdb, p_a, q_thermogenesis_total)
    assert res_sh == 0.0



import numpy as np
import pytest

from pythermalcomfort.classes_return import GaggeTwoNodesSleep
from pythermalcomfort.models import two_nodes_gagge_sleep


def test_two_nodes_gagge_sleep_single_input() -> None:
    """Test the two_nodes_gagge_sleep function with scalar inputs."""
    result = two_nodes_gagge_sleep(18, 18, 0.05, 50, 1.4, thickness_quilt=1.76)

    # expected outputs
    expected = {
        "set": np.array([24.28]),
        "t_core": np.array([37.03]),
        "t_skin": np.array([33.67]),
        "wet": np.array([0.27]),
        "t_sens": np.array([1.12]),
        "disc": np.array([1.47]),
        "e_skin": np.array([28.75]),
        "met_shivering": np.array([0.0]),
        "alfa": np.array([0.13]),
        "skin_blood_flow": np.array([7.11]),
    }

    # compare each field with a reasonable tolerance
    for field, exp in expected.items():
        actual = getattr(result, field)
        np.testing.assert_allclose(actual, exp, atol=0.01, rtol=0.005)


def test_two_nodes_gagge_sleep_long_duration() -> None:
    """Test the two_nodes_gagge_sleep function with a longer duration input."""
    duration = 481
    ta = np.repeat(18, duration)
    tr = np.repeat(18, duration)
    vel = np.repeat(0.05, duration)
    rh = np.repeat(50, duration)
    clo_a = np.repeat(1.4, duration)
    thickness1 = np.repeat(1.76, duration)

    result = two_nodes_gagge_sleep(ta, tr, vel, rh, clo_a, thickness1)

    # Assert return type and shape
    assert isinstance(result, GaggeTwoNodesSleep)
    assert result.set.shape == (duration,)
    assert result.t_core.shape == (duration,)
    assert result.t_skin.shape == (duration,)

    # for field in fields:
    #     print(f"{getattr(result, field)[-1]:.2f},")

    first_row_expected = [
        24.29,
        37.03,
        33.67,
        0.27,
        1.13,
        1.48,
        28.76,
        0.00,
        0.14,
        7.11,
    ]
    last_row_expected = [
        22.23,
        36.23,
        31.06,
        0.06,
        -0.73,
        -0.73,
        5.21,
        0.00,
        0.25,
        2.98,
    ]

    fields = [
        "set",
        "t_core",
        "t_skin",
        "wet",
        "t_sens",
        "disc",
        "e_skin",
        "met_shivering",
        "alfa",
        "skin_blood_flow",
    ]

    # check first row
    for field, exp in zip(fields, first_row_expected, strict=False):
        np.testing.assert_allclose(
            getattr(result, field)[0],
            exp,
            atol=0.01,
            rtol=0.005,
            err_msg=f"first {field} mismatch",
        )

    # check last row
    for field, exp in zip(fields, last_row_expected, strict=False):
        np.testing.assert_allclose(
            getattr(result, field)[-1],
            exp,
            atol=0.01,
            rtol=0.005,
            err_msg=f"last {field} mismatch",
        )


def test_length_mismatch_raises_value_error() -> None:
    """Test that length mismatch in input lists raises ValueError."""
    with pytest.raises(ValueError) as exc:
        two_nodes_gagge_sleep(
            [18, 18],
            [18],  # length mismatch
            [0.05, 0.05],
            [50, 50],
            [1.4, 1.4],
            [1.76, 1.76],
        )

    assert "must have the same length" in str(exc.value)


def test_unexpected_kwargs_raises_type_error() -> None:
    """Test that unexpected keyword arguments raise TypeError."""
    with pytest.raises(TypeError) as exc:
        two_nodes_gagge_sleep(18, 18, 0.05, 50, 1.4, 1.76, foo=123)
    assert "Unexpected keyword arguments" in str(exc.value)


def test_invalid_kwarg_type_raises_type_error() -> None:
    """Test that a non-numeric type for tdb raises TypeError."""
    with pytest.raises(TypeError) as exc:
        two_nodes_gagge_sleep("string", 18, 0.05, 50, 1.4, 1.76)
    msg = str(exc.value)
    assert "tdb" in msg


def test_tickness_quilt_negative() -> None:
    """Test that a negative thickness_quilt raises ValueError."""
    with pytest.raises(ValueError):
        two_nodes_gagge_sleep(18, 18, 0.05, 50, 1.4, thickness_quilt=-1.76)



from pythermalcomfort.models import at
from tests.conftest import Urls, is_equal, retrieve_reference_table, validate_result


def test_at(get_test_url, retrieve_data) -> None:
    """Test that the function calculates the AT correctly for various inputs."""
    reference_table = retrieve_reference_table(
        get_test_url,
        retrieve_data,
        Urls.AT.name,
    )
    tolerance = reference_table["tolerance"]

    for entry in reference_table["data"]:
        inputs = entry["inputs"]
        outputs = entry["outputs"]
        result = at(**inputs)

        validate_result(result, outputs, tolerance)


def test_at_list_input() -> None:
    """Test that the function calculates the AT correctly for list inputs."""
    result = at([25, 25, 25], [30, 30, 30], [0.1, 0.1, 0.1])
    is_equal(result, [24.1, 24.1, 24.1], 0.1)


def test_at_q() -> None:
    """Test that the function calculates the AT correctly for given inputs."""
    result = at(25, 30, 0.1, 100)
    is_equal(result, 25.3, 0.1)
